import warnings

from v0.relationalai.semantics.lqp.types import lqp_type_to_sql, meta_type_to_lqp, is_numeric

__all__ = ['lqp_type_to_sql', 'meta_type_to_lqp', 'is_numeric']

warnings.warn(
    "relationalai.early_access.lqp.types is deprecated, "
    "Please migrate to relationalai.semantics.lqp.types",
    DeprecationWarning,
    stacklevel=2,
)


from io import StringIO
from typing import Optional, Hashable

import numpy as np
import pandas as pd

from v0.relationalai.semantics import define, where
from v0.relationalai.semantics.std import decimals
from v0.relationalai.semantics.internal import internal as b
from v0.relationalai.early_access.dsl.bindings.common import BindableColumn, AbstractBindableTable
from v0.relationalai.early_access.dsl.snow.common import CsvColumnMetadata
from v0.relationalai.early_access.dsl.utils import normalize


class BindableCsvColumn(BindableColumn, b.Relationship):
    _metadata: CsvColumnMetadata
    _column_basic_type: str

    def __init__(self, metadata: CsvColumnMetadata, source: 'CsvTable', model):
        BindableColumn.__init__(self, metadata.name, metadata.datatype, source, model)
        fqn, fields, field_refs = self._relationship_ctor_inputs(metadata, source)
        b.Relationship.__init__(self, fqn, model=model.qb_model(), fields=fields, field_refs=field_refs)
        self._metadata = metadata
        self._column_basic_type = "Int64" if metadata.datatype._name == b.Integer._name else "string"

    @staticmethod
    def _relationship_ctor_inputs(metadata: CsvColumnMetadata, source: 'CsvTable'):
        """
        Handles the construction of the Relationship name components.

        We use a separate QB relationship per source column, as this allows proper typing.
        """
        col_name = metadata.name
        col_type = metadata.datatype
        source_name = source.physical_name()
        fqn = f"{source_name}_{col_name}"
        fields = [source._row_field, b.Field(name=col_name, type_str=str(col_type), type=col_type)]
        field_refs = [source._ref, col_type.ref()]
        return fqn, fields, field_refs

    def __call__(self, *args):
        """
        Allows the column to be called as a function, which is handy for manual data weaving.

        Example:
            row = Integer.ref()
            where(
                {source}.ID(row, id),
                {source}.NAME(row, name),
                person := Person.new(id=id)
            ).define(
                person,
                Person.name(person, name)
            )
        """
        if len(args) != 2:
            raise ValueError(f'Expected 2 arguments passed to a call to BindableColumn, got {len(args)}')
        return b.Relationship.__call__(self, *args)

    @property
    def metadata(self):
        return self._metadata

    def basic_type(self):
        return self._column_basic_type

    def decimal_scale(self) -> Optional[int]:
        typ = self.type()
        if decimals.is_decimal(typ):
            return decimals.scale(typ)
        else:
            return None

    def __repr__(self):
        return f"CSV:{self._source.physical_name()}.{self.physical_name()}"


class CsvTable(AbstractBindableTable[BindableCsvColumn]):
    _basic_type_schema: dict[Hashable, str]
    _csv_data: list[pd.DataFrame]
    _num_rows: int

    def __init__(self, fqn: str, schema: dict[str, b.Concept], model):
        AbstractBindableTable.__init__(self, fqn, model, set())
        self._initialize(schema, model)

    def _initialize(self, schema: dict[str, b.Concept], model):
        self._concept = b.Concept(self._fqn, extends=[b.Concept.builtins["RowId"]])
        b.Concept.globals[self._fqn.lower()] = self._concept
        ref = self._concept.ref("row_id")
        assert isinstance(ref, b.Ref)
        self._ref = ref
        self._row_field = b.Field(name="row_id", type_str=self._fqn, type=self._concept)
        self._csv_data = list()
        self._num_rows = 0
        self._cols = {column_name: BindableCsvColumn(CsvColumnMetadata(column_name, column_type), self, model)
                      for column_name, column_type in schema.items()}
        self._basic_type_schema = {col.metadata.name: col.basic_type() for col in self._cols.values()}

    def _compile_lookup(self, compiler:b.Compiler, ctx:b.CompilerContext):
        return compiler.lookup(self._ref, ctx)

    def _to_keys(self) -> list[b.Ref]:
        assert isinstance( self._ref, b.Ref)
        return [self._ref]

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self.physical_name() + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype}' for _, col in self._cols.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    @property
    def csv_data(self) -> list[pd.DataFrame]:
        return self._csv_data

    def physical_name(self) -> str:
        return self._fqn.lower()

    def data(self, csv_data: str):
        csv_df = pd.read_csv(StringIO(normalize(csv_data)), dtype=self._basic_type_schema)
        self._csv_data.append(csv_df)
        CsvSourceModule.generate(self, csv_df, row_offset=self._num_rows)
        # update offset now that we generated the rules
        self._num_rows += len(csv_df)

class CsvSourceModule:

    @staticmethod
    def generate(table: CsvTable, data: pd.DataFrame, row_offset: int = 0):
        for local_index, row in enumerate(data.itertuples(index=False)):
            row_index = row_offset + local_index
            for column_name in data.columns:
                value = getattr(row, column_name)
                if pd.notna(value):
                    column = table.__getattr__(column_name)
                    column_type = column.type()
                    if column_type._name == b.Date._name:
                        CsvSourceModule._row_to_date_value_rule(column, row_index, value)
                    elif column_type._name == b.DateTime._name:
                        CsvSourceModule._row_to_date_time_value_rule(column, row_index, value)
                    elif b.is_decimal(column_type):
                        CsvSourceModule._row_to_decimal_value_rule(column, row_index, value)
                    else:
                        CsvSourceModule._row_to_value_rule(column, row_index, value)

    @staticmethod
    def _row_to_value_rule(column, row, value):
        # if numpy scalar, convert to a native Python type
        if isinstance(value, np.generic):
            value = value.item()
        define(column(row, value))

    @staticmethod
    def _row_to_date_value_rule(column, row, value):
        parse_date = b.Relationship.builtins['parse_date']
        rez = b.Date.ref()
        where(parse_date(value, 'yyyy-mm-dd', rez)).define(column(row, rez))

    @staticmethod
    def _row_to_date_time_value_rule(column, row, value):
        parse_datetime = b.Relationship.builtins['parse_datetime']
        rez = b.DateTime.ref()
        where(parse_datetime(value, 'yyy-mm-dd HH:MM:SS z', rez)).define(column(row, rez))

    @staticmethod
    def _row_to_decimal_value_rule(column, row, value):
        define(column(row, decimals.parse(value, column.type())))



from v0.relationalai.semantics.metamodel.rewrite import Flatten, \
    DNFUnionSplitter, ExtractNestedLogicals, flatten
from v0.relationalai.semantics.lqp.rewrite import Splinter,  \
    ExtractKeys, FunctionAnnotations

__all__ = ["Splinter", "Flatten", "DNFUnionSplitter", "ExtractKeys",
           "ExtractNestedLogicals", "FunctionAnnotations", "flatten"]



"""
Nested configuration field models for the multi-connection config system.

This module defines nested config models:
- EngineConfig: Engine execution settings
- DataConfig: Data loading and streaming settings
- CompilerConfig: Compiler-specific settings
- ModelConfig: Model execution settings
- ReasonerConfig: Reasoner execution settings (including nested ReasonerRuleConfig)
- DebugConfig: Debug server settings

These models are used in Config (with default instances) and Profile (as optional overrides).
"""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field


class EngineConfig(BaseModel):
    name: str | None = Field(default=None, description="Engine name for execution")
    size: str | None = Field(default=None, description="Size of the engine")
    auto_suspend_mins: int | None = Field(default=None, description="Auto-suspend engine after N minutes")
    show_all_sizes: bool = Field(
        default=False,
        description="Show all available engine sizes"
    )


class DataConfig(BaseModel):
    wait_for_stream_sync: bool = Field(default=True, description="Wait for stream synchronization before processing")
    ensure_change_tracking: bool = Field(default=False, description="Enable change tracking for data modifications")
    data_freshness_mins: int | None = Field(default=None, description="Data freshness timeout in minutes")
    query_timeout_mins: int | None = Field(
        default=None,
        description="Query timeout in minutes - aborts queries that exceed this duration"
    )
    download_url_type: Literal["internal", "external"] | None = Field(
        default=None,
        description="Type of download URL for data exports (internal or external)"
    )
    check_column_types: bool = Field(
        default=True,
        description="Check column types during data loading"
    )


class CompilerConfig(BaseModel):
    use_monotype_operators: bool = Field(default=False, description="Use monotype operators in compilation")
    show_corerel_errors: bool = Field(default=True, description="Show CoreRel error messages")
    dry_run: bool = Field(default=False, description="Run compilation in dry-run mode")
    inspect_df: bool = Field(default=False, description="Inspect DataFrame during compilation")
    use_value_types: bool = Field(default=False, description="Use value types in compilation")
    debug_hidden_keys: bool = Field(default=False, description="Debug hidden keys in compilation")
    wide_outputs: bool = Field(default=False, description="Use wide output format for query results")
    strict: bool = Field(default=False, description="Enable strict validation mode")

    # Experimental compiler optimizations
    use_inlined_intermediates: bool = Field(default=False, description="Use inlined intermediate results")
    inline_value_maps: bool = Field(default=False, description="Inline value maps in weaver")
    inline_entity_maps: bool = Field(default=False, description="Inline entity maps in weaver")


class ModelConfig(BaseModel):
    keep: bool = Field(default=False, description="Keep model after execution")
    isolated: bool = Field(default=True, description="Run model in isolated mode")
    nowait_durable: bool = Field(default=True, description="Don't wait for durable operations")


class ReasonerRuleConfig(BaseModel):
    use_lqp: bool = Field(default=True, description="Use LQP for reasoner rule execution")


class ReasonerConfig(BaseModel):
    rule: ReasonerRuleConfig = Field(default_factory=ReasonerRuleConfig)
    use_sql: bool = Field(default=False, description="Use SQL execution instead of LQP")


class DebugConfig(BaseModel):
    enabled: bool = Field(default=True, description="Enable debug mode")
    host: str | None = Field(default=None, description="Debug server host")
    port: int = Field(default=8080, description="Debug server port")
    show_debug_logs: bool = Field(default=False, description="Show debug log messages")
    show_full_traces: bool = Field(default=False, description="Show full stack traces in errors")



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional
from v0.relationalai.semantics.metamodel import ir, factory as f, helpers, visitor
from v0.relationalai.semantics.metamodel.compiler import Pass
from v0.relationalai.semantics.metamodel.util import OrderedSet, ordered_set
from v0.relationalai.semantics.metamodel import dependency
from v0.relationalai.semantics.metamodel import builtins

from functools import reduce

class ExtractCommon(Pass):
    """
    Pass to analyze Logical bodies and extract lookups in their own Logical if it makes
    sense. The heuristic is that, if there are multiple lookups, and there are also multiple
    sibling nested logicals that will eventually be extracted by Flatten, then it makes
    sense to extract these logicals into their own "rule", and then make the original body
    just lookup this common rule.

    From:
        Logical
            Logical
                lookup1
                lookup2
                Logical1 ...
                Logical2 ...
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive common
            Logical
                lookup common
                Logical1 ...
                Logical2 ...
    """

    # The extraction plan heuristic is as follows:
    #
    # Given a set of binder tasks B and a set of extractable tasks E, we find:
    #   - A subset of common tasks C in B, and
    #   - A subset of exposed variables V output from tasks in C
    # where:
    #   - The intersection of common dependencies of all tasks in E are contained in C
    #     (including transitive dependencies)
    #   - The union of input variables for all tasks in E intersected with the output
    #     variables of tasks in C are contained in V

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # create the  context
        ctx = ExtractCommon.Context(model, options)

        # rewrite the root
        replacement = self.handle(model.root, ctx)

        # the new root contains the extracted top level logicals and the rewritten root
        if ctx.rewrite_ctx.top_level:
            new_root = ir.Logical(model.root.engine, tuple(), tuple(ctx.rewrite_ctx.top_level + [replacement]))
        else:
            new_root = replacement

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen(),
            model.types,
            new_root
        )

    #--------------------------------------------------
    # Extra classes
    #--------------------------------------------------

    class Context():
        def __init__(self, model: ir.Model, options: dict):
            self.rewrite_ctx = helpers.RewriteContext()
            self.info = dependency.analyze(model.root)
            self.options = options

    @dataclass
    class ExtractionPlan():
        # tasks to extract to the body of the common logical
        common_body: OrderedSet[ir.Task]
        # tasks to remain in the original body
        remaining_body: OrderedSet[ir.Task]
        # variables to be exposed by the common logical
        exposed_vars: OrderedSet[ir.Var]
        # map from nested composite to the tasks in the common body that still need to be
        # included in its body, because it contains variables not exposed by the common logical
        local_dependencies: dict[ir.Task, OrderedSet[ir.Task]]
        # a reference to the common connection created for this plan, if any
        common_reference: Optional[ir.Lookup] = None

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    def handle(self, task: ir.Task, ctx: Context):
        # Currently we only extract if it's a sequence of Logicals, but we could in the
        # future support other intermediate nodes
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        else:
            return task

    def handle_logical(self, task: ir.Logical, ctx: Context):
        # Process the original body to find candidates for extraction. Each task is in one
        # of three categories:
        # - Binders: tasks that bind variables. These are candidates for extracting into
        #   the common body.
        # - Flattenables: tasks that will later be extracted by the Flatten pass
        # - Other: tasks that are neither binders nor flattenables; these will remain
        #   in the body as-is.

        binders = ordered_set()
        flattenables = ordered_set()
        other = ordered_set()

        for child in task.body:
            if _is_binder(child):
                binders.add(child)
            elif _is_flattenable(ctx, child):
                flattenables.add(child)
            else:
                other.add(child)

        # The new body of the rewritten task
        body:OrderedSet[ir.Task] = ordered_set()

        # Quick check to see if it's worth doing more analysis; we only want to extract
        # common binders if there are multiple, and there are also multiple flattenables
        # that will be extracted by the flatten pass later (so that they can share the
        # extracted logic).
        plan: Optional[ExtractCommon.ExtractionPlan] = None
        if len(binders) > 1 and len(flattenables) > 1:
            plan = _create_extraction_plan(ctx, binders, flattenables, other)
            if plan:
                # plan is worthwhile, extract the common body and add the connection to the body
                exposed_vars = plan.exposed_vars.get_list()
                plan.common_reference = f.lookup(helpers.extract(task, plan.common_body, exposed_vars, ctx.rewrite_ctx, "common"), exposed_vars)

                # Add plan common reference to the body.
                body.add(plan.common_reference)

        # recursively handle children
        for child in task.body:
            # skip children that were extracted
            if plan and child not in other and child not in plan.remaining_body and child not in flattenables:
                continue

            # no plan or child is not a composite, so just add the handled to the body
            if not plan or child not in flattenables:
                body.add(self.handle(child, ctx))
                continue

            # there is a plan and the child is in composites, so...
            replacement = self.handle(child, ctx)

            # this child needs either extra local dependencies or the common reference
            if child in plan.local_dependencies:
                # the new body will have maybe the common reference and the local deps
                replacement_body = ordered_set()

                hoisted = OrderedSet()
                if isinstance(replacement, ir.Logical):
                    # if replacement is a logical, just keep the same hoisted vars
                    hoisted.update(replacement.hoisted)
                else:
                    # otherwise, we need to hoist the vars that are output from local deps
                    # and input to the replacement task
                    dep_outputs = OrderedSet()
                    for d in plan.local_dependencies.get(child, ordered_set()):
                        dep_outputs.update(ctx.info.task_outputs(d))
                    hoisted.update(dep_outputs & ctx.info.task_inputs(replacement))

                if child in plan.local_dependencies:
                    for local_dep in plan.local_dependencies[child]:
                        replacement_body.add(local_dep.clone())

                if isinstance(replacement, ir.Logical):
                    # if the replacements is a logical, we can just add to the body
                    body.add(replacement.reconstruct(
                        replacement.engine,
                        tuple(hoisted.get_list()),
                        tuple(replacement_body.update(replacement.body).get_list()),
                        replacement.annotations
                    ))
                else:
                    # Otherwise, wrap the local dependencies in a Lookup where the output
                    # variables are hoisted, and keep the computed replacement.
                    body.add(f.logical(replacement_body.get_list(), hoisted.get_list(), replacement.engine))
                    body.add(replacement)
            else:
                # child does not need extras in the body, just add it to the main body
                body.add(replacement)

        return ir.Logical(task.engine, task.hoisted, tuple(body))


#--------------------------------------------------
# Utilities
#--------------------------------------------------

def _create_extraction_plan(ctx: ExtractCommon.Context, binders: OrderedSet[ir.Task], flattenables: OrderedSet[ir.Task], others: OrderedSet[ir.Task]) -> Optional[ExtractCommon.ExtractionPlan]:
    """
    Compute a plan to extract tasks in this frame that are common dependencies
    across these composite tasks.
    """
    # If there are any pragma lookups, then don't extract anything. Pragma lookups are
    # designed to control execution order, and extracting them may affect their
    # semantics.
    for b in binders:
        if isinstance(b, ir.Lookup) and builtins.is_pragma(b.relation):
            return None

    # Compute intersection of task dependencies
    all_deps = [ctx.info.task_dependencies(f) for f in flattenables]
    deps = reduce(lambda a, b: a & b, all_deps)
    common_body = binders & deps

    # We don't need to extract anything if there's only zero or one common tasks
    if len(common_body) < 2:
        return None

    # Keep track of remaining tasks that are not extracted in the common body
    remaining = ordered_set()

    # Compute the vars that should be output from the common body. These are the union of
    # all input vars across all non-extracted tasks, intersected with output vars of
    # the common body.

    # First, compute the output vars of the common body
    common_body_output_vars = OrderedSet()
    for child in common_body:
        common_body_output_vars.update(ctx.info.task_outputs(child))

    # Next, compute the union of the input vars of all non-extracted tasks
    non_extracted_tasks = (binders - common_body) | flattenables | others
    all_exposed_vars: list[OrderedSet[ir.Var]] = []
    for t in non_extracted_tasks:
        input_vars = ctx.info.task_inputs(t)
        all_exposed_vars.append(input_vars if input_vars else OrderedSet())

    exposed_vars = reduce(lambda a, b: a | b, all_exposed_vars) & common_body_output_vars

    # If there are no vars in common, then it's not worth extracting
    if not exposed_vars:
        return None

    # Make sure that all local dependencies of the common body are included in the common
    # body. This is important for the safety of this rewrite.
    for task in common_body:
        local_deps = ctx.info.local_dependencies(task)
        if local_deps:
            common_body.update(local_deps & binders)

    # check which of the original binders remain, and make sure their dependencies also stay
    for binder in binders:
        if binder not in common_body:
            remaining.add(binder)
            deps = _compute_local_dependencies(ctx, binders, binder, exposed_vars)
            if deps:
                remaining.update(deps)

    # for each composite, check if there are additional tasks needed, because the task
    # depends on it but it is not exposed by the vars
    local_dependencies: dict[ir.Task, OrderedSet[ir.Task]] = dict()
    for flattenable in flattenables:
        local = _compute_local_dependencies(ctx, binders, flattenable, exposed_vars)
        if local:
            local_dependencies[flattenable] = local

    return ExtractCommon.ExtractionPlan(common_body, remaining, exposed_vars, local_dependencies)

def _compute_local_dependencies(ctx: ExtractCommon.Context, binders: OrderedSet[ir.Task], composite: ir.Task, exposed_vars: OrderedSet[ir.Var]):
    """
    The tasks in common_body will be extracted into a logical that will expose the exposed_vars.
    Compute which additional dependencies are needed specifically for this composite, because
    it depends on some tasks that are extracted to common_body but not exposed by exposed_vars.
    """

    # working list of vars we still need to fulfill
    inputs = ctx.info.task_inputs(composite)
    if not inputs:
        return None

    # vars exposed by exposed vars + tasks added to the local body
    vars_exposed = OrderedSet.from_iterable(exposed_vars)
    vars_needed = (inputs - vars_exposed)
    if not vars_needed:
        return None

    # this is a greedy algorithm that uses the first task in the common body that provides
    # a variable needed; it may result in sub-optimal extraction, but should be correct
    local_body = ordered_set()
    while(vars_needed):
        v = vars_needed.pop()
        for x in binders:
            if x not in local_body:
                # an x that is not yet in local_body can fulfill v
                x_outputs = ctx.info.task_outputs(x)
                if x_outputs and v in x_outputs:
                    # add it to local_body and add its outputs to vars exposed
                    local_body.add(x)
                    vars_exposed.add(x_outputs)
                    # but add its inputs the vars now needed
                    inputs = ctx.info.task_inputs(x)
                    if inputs:
                        vars_needed.update(inputs - vars_exposed)
    return local_body

def _is_binder(task: ir.Task):
    # If the task itself is a binder
    if any(isinstance(task, binder) for binder in (ir.Lookup, ir.Construct, ir.Exists, ir.Data, ir.Not)):
        return True

    # If the task is a Logical containing only binders
    if isinstance(task, ir.Logical) and all(_is_binder(c) for c in task.body):
        return True

    # If the task is a Union containing only binders
    if isinstance(task, ir.Union) and all(_is_binder(c) for c in task.tasks):
        return True

    return False

def _is_flattenable(ctx: ExtractCommon.Context, task: ir.Task):
    # Each output will be flattened into its own top-level def
    if isinstance(task, ir.Output):
        return True

    extractable_types = (ir.Update, ir.Aggregate, ir.Match, ir.Rank)
    return isinstance(task, ir.Logical) and len(visitor.collect_by_type(extractable_types, task)) > 0



import json
import typing
from collections import defaultdict
from typing import TypeVar

from v0.relationalai.early_access.dsl import Model, ExternalRelation, ValueType, ValueSubtype, Relationship
from v0.relationalai.early_access.dsl.core.types import Type
from v0.relationalai.early_access.dsl.core.types.standard import standard_value_types
from v0.relationalai.early_access.dsl.ontologies.constraints import Mandatory, Unique, RoleValueConstraint
from v0.relationalai.early_access.dsl.ontologies.roles import AbstractRole
from v0.relationalai.early_access.dsl.ontologies.subtyping import SubtypeConstraint, SubtypeArrow, \
    ExclusiveSubtypeConstraint, InclusiveSubtypeConstraint
from v0.relationalai.early_access.dsl.types.concepts import Concept
from v0.relationalai.semantics.metamodel.util import Printer, OrderedSet

# Define a generic type variable constrained to SubtypeConstraint or its subclasses
T = TypeVar("T", bound=SubtypeConstraint)

class PythonPrinter(Printer):
    _mandatory_roles: 'OrderedSet[AbstractRole]'
    _single_unique_roles: 'OrderedSet[AbstractRole]'
    _composite_preferred_identifiers: 'OrderedSet[Unique]'
    _internal_preferred_identifier_roles: 'OrderedSet[AbstractRole]'
    _relationship_uniqueness_constraints: 'dict[Relationship, OrderedSet[Unique]]'

    def __init__(self, io: typing.Optional[typing.IO[str]] = None):
        # Set the base class field (frozen)
        object.__setattr__(self, 'io', io)

        # Set all mutable fields using object.__setattr__
        object.__setattr__(self, '_mandatory_roles', OrderedSet())
        object.__setattr__(self, '_single_unique_roles', OrderedSet())
        object.__setattr__(self, '_composite_preferred_identifiers', OrderedSet())
        object.__setattr__(self, '_internal_preferred_identifier_roles', OrderedSet())
        object.__setattr__(self, '_relationship_uniqueness_constraints', defaultdict(OrderedSet))

    def to_python_string(self, model: Model) -> None:
        self._process_constraints(model)

        self._print_nl("import relationalai.early_access.dsl as rai")
        self._nl()

        self._print_nl(f"model = rai.Model(name='{model.name}', is_primary={model.is_primary})")
        self._print_nl("Concept = model.concept")
        self._print_nl("ValueType = model.value_type")
        self._print_nl("ValueSubType = model.value_sub_type")
        self._print_nl("EntityType = model.entity_type")
        self._print_nl("SubtypeArrow = model.subtype_arrow")
        self._print_nl("Relationship = model.relationship")
        self._print_nl("RefScheme = model.ref_scheme")
        self._print_nl("ExternalRelation = model.external_relation")
        self._print_nl("Query = model.query")
        self._print_nl("RoleValueConstraint = model.role_value_constraint")
        self._print_nl("CsvTable = model.csv_table")

        self._nl()
        self._handle_value_types(model)

        self._nl()
        self._handle_entity_types(model)

        self._nl()
        self._handle_relationships(model)

        self._handle_composite_reference_schemas()

        self._nl()
        self._handle_subtype_arrows(model)

        self._nl()
        self._handle_external_relations(model)

        self._nl()
        self._handle_role_value_constraints(model)

    def _handle_value_types(self, model: Model) -> None:
        for vt in model.value_types():
            name = vt.display()
            if isinstance(vt, ValueType):
               self._print_nl(f"{name} = ValueType('{name}', {', '.join(self._get_type(t) for t in vt._types)})")
            elif isinstance(vt, ValueSubtype) and vt.parent():
                self._print_nl(f"{name} = ValueSubType('{name}', {self._get_type(vt.parent())})")
            elif isinstance(vt, Concept):
                self._print_nl(f"{name} = ValueType('{name}', {', '.join(self._get_type(t) for t in vt._types)})")

    def _handle_entity_types(self, model: Model) -> None:
        for name, et in model.entity_types_map().items():
            self._print_nl(f"{name} = EntityType('{name}'{self._print_if_not_empty('ref_mode', et.ref_schema_name())})")

    def _handle_relationships(self, model: Model) -> None:
        for rel in model.relationships():
            if not rel.is_subtype():
                self._print_nl("with Relationship() as rel:")
                self._handle_relationship_roles(rel)
                self._handle_relationship_relations(rel)
                self._handle_relationship_uniqueness_constraints(rel, self._relationship_uniqueness_constraints.get(rel, OrderedSet()))
                self._nl()

    def _handle_relationship_roles(self, rel):
        for i, r in enumerate(rel.roles()):
            self._indent_print_nl(1, f"rel.role({self._get_type(r.player())}"
                                     f"{self._print_if_not_empty('name', r.name())}"
                                     f"{self._print_if_true('unique', r in self._single_unique_roles)}"
                                     f"{self._print_if_true('mandatory', r in self._mandatory_roles)}"
                                     f"{self._print_if_true('primary_key', r in self._internal_preferred_identifier_roles)})")
            if r.prefix or r.postfix:
                elements = list(filter(None, [
                    self._print_first_if_not_empty('prefix', r.prefix),
                    self._print_first_if_not_empty('postfix', r.postfix)
                ]))
                self._indent_print_nl(1, f"rel.role_at({i}).verbalization({', '.join(elements)})")

    def _handle_relationship_relations(self, rel):
        for relation in rel.relations():
            reading = relation.reading()
            if reading:

                num_roles = len(reading.roles)
                num_texts = len(reading.text_frags)

                elements = []

                for i in range(num_texts):
                    role = reading.role_at(i)
                    elements.append(f"rel.role_at({rel.role_index(role)})")  # Role first
                    elements.append(f"'{reading.text_frags[i]}'")  # Then text

                if num_roles > num_texts:
                    role = reading.role_at(num_texts)
                    elements.append(f"rel.role_at({rel.role_index(role)})")  # Add last role if needed

                self._indent_print_nl(1, f"rel.relation({', '.join(elements)}"
                                         f"{self._print_if_not_empty('name', reading.rel_name)}"
                                         f"{self._print_if_true('functional', relation.signature().functional())})")

    def _handle_relationship_uniqueness_constraints(self, rel, constraints: OrderedSet[Unique]) -> None:
        for c in constraints:
            elements = [f"rel.role_at({rel.role_index(role)})" for role in c.roles()]
            self._indent_print_nl(1, f"rel.unique({', '.join(elements)})")

    def _handle_composite_reference_schemas(self) -> None:
        for preferred_id in self._composite_preferred_identifiers:

            elements = []

            for role in preferred_id.roles():
                relationship = role.part_of

                relation = self._lookup_relation_by_second_role(relationship, role)

                player_name = self._get_type(relation.first())
                rel_name = relation.rel_name()
                elements.append(f"{player_name}.{rel_name}")

            self._print_nl(f"RefScheme({', '.join(elements)})")

    def _handle_subtype_arrows(self, model:Model) -> None:
        subtype_arrows_by_type = self._group_subtype_arrows_by_type(model)
        inclusive_subtype_constraints_by_type = self._get_inclusive_subtype_constraints_by_type(model)
        exclusive_subtype_constraints_by_type = self._get_exclusive_subtype_constraints_by_type(model)

        for et in model.entity_types():

            subtype_arrows = subtype_arrows_by_type.get(et, OrderedSet())
            inclusive_subtype_constraints = inclusive_subtype_constraints_by_type.get(et, OrderedSet())
            exclusive_subtype_constraints = exclusive_subtype_constraints_by_type.get(et, OrderedSet())

            name = et.display()

            # Common elements in both sets
            common_constraints = [i for i in inclusive_subtype_constraints
                                    for e in exclusive_subtype_constraints
                                    if self.constraints_equal(i, e)]
            if len(common_constraints) > 0:
                for c in common_constraints:
                    self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in c.arrows)}], "
                                   f"exclusive=True, inclusive=True)")

            # Only in inclusive (but not in exclusive)
            only_inclusive_constraints = [i for i in inclusive_subtype_constraints
                                            if not any(self.constraints_equal(i, e) for e in exclusive_subtype_constraints)]
            if len(only_inclusive_constraints) > 0:
                for c in only_inclusive_constraints:
                    self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in c.arrows)}], "
                                   f"inclusive=True)")

            # Only in exclusive (but not in inclusive)
            only_exclusive_constraints = [e for e in exclusive_subtype_constraints
                                            if not any(self.constraints_equal(e, i) for i in inclusive_subtype_constraints)]
            if len(only_exclusive_constraints) > 0:
                for c in only_exclusive_constraints:
                    self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in c.arrows)}], "
                                   f"exclusive=True)")

            # Get all arrows from both inclusive and exclusive constraints
            all_arrows_in_constraints = OrderedSet.from_iterable(a for c in inclusive_subtype_constraints for a in c.arrows) | \
                                        OrderedSet.from_iterable(a for c in exclusive_subtype_constraints for a in c.arrows)

            # Subtract the arrows in constraints from subtype_arrows
            remaining_subtype_arrows = subtype_arrows - all_arrows_in_constraints
            if len(remaining_subtype_arrows) > 0:
                self._print_nl(f"SubtypeArrow({name}, [{', '.join(self._get_type(a.start) for a in remaining_subtype_arrows)}])")

    def _handle_external_relations(self, model: Model) -> None:
        for relation in self._get_external_relations(model):
            name = relation.rel_name()
            type_args = ', '.join(self._get_type(t) for t in relation.signature().types())

            if name == "output":
                self._print_nl(f"Query({type_args})")
            else:
                self._print_nl(f"ExternalRelation('{name}', {type_args})")

    def _handle_role_value_constraints(self, model: Model) -> None:
        for c in self._get_role_value_constraints(model):
            role = c.role()
            relationship = role.part_of

            relation = self._lookup_relation_by_second_role(relationship, role)

            player_name = self._get_type(relation.first())
            rel_name = relation.rel_name()

            self._print_nl(f"RoleValueConstraint({player_name}.{rel_name}, {json.dumps(c.values())})")

    def _process_constraints(self, model: Model) -> None:
        for c in model.constraints():
            if isinstance(c, Mandatory):
                self._mandatory_roles.add(c.role)
            elif isinstance(c, Unique):
                if c.is_preferred_identifier:
                    if len(c.roles()) == 1:
                        self._internal_preferred_identifier_roles.add(c.roles()[0])
                    else:
                        self._composite_preferred_identifiers.add(c)
                elif len(c.roles()) == 1:
                    self._single_unique_roles.add(c.roles()[0])
                else:
                    self._relationship_uniqueness_constraints[c.roles()[0].part_of].add(c)

    @staticmethod
    def _group_subtype_arrows_by_type(model: Model) -> dict[Type, OrderedSet[SubtypeArrow]]:
        subtype_arrows_by_type = defaultdict(OrderedSet)
        for a in model.subtype_arrows():
            subtype_arrows_by_type[a.end].add(a)
        return dict(subtype_arrows_by_type)

    @staticmethod
    def _get_subtype_constraints_by_type(model: Model, constraint_type: typing.Type[T]) -> dict[Type, OrderedSet[SubtypeConstraint]]:
        constraints_by_type: dict[Type, OrderedSet[SubtypeConstraint]] = defaultdict(OrderedSet)

        for c in model.subtype_constraints():
            if isinstance(c, constraint_type):
                for a in c.arrows:
                    constraints_by_type[a.end].add(c)

        return dict(constraints_by_type)

    def _get_exclusive_subtype_constraints_by_type(self, model: Model) -> dict[Type, OrderedSet[SubtypeConstraint]]:
        return self._get_subtype_constraints_by_type(model, ExclusiveSubtypeConstraint)

    def _get_inclusive_subtype_constraints_by_type(self, model: Model) -> dict[Type, OrderedSet[SubtypeConstraint]]:
        return self._get_subtype_constraints_by_type(model, InclusiveSubtypeConstraint)

    @staticmethod
    def _get_role_value_constraints(model: Model) -> OrderedSet[RoleValueConstraint]:
        return OrderedSet.from_iterable(c for c in model.constraints() if isinstance(c, RoleValueConstraint))

    @staticmethod
    def _get_external_relations(model: Model) -> OrderedSet[ExternalRelation]:
        return OrderedSet.from_iterable(r for r in model.relations() if isinstance(r, ExternalRelation))

    @staticmethod
    def _lookup_relation_by_second_role(relationship, role):
        # Find the matching relation where the 2nd role is `role`
        relation = next(
            (rel for rel in relationship.relations() if rel.reading().roles[1] == role),
            None
        )
        if relation is None:
            raise Exception(f"Could not find matching relation for role player {role.player().name()} "
                            f"in relationship {relationship._name()}")
        return relation

    @staticmethod
    def constraints_equal(a: SubtypeConstraint, b: SubtypeConstraint) -> bool:
        return frozenset(a.arrows) == frozenset(b.arrows)

    @staticmethod
    def _get_type(t: Type) -> str:
        return f"rai.{t.display()}" if t.display() in standard_value_types else t.display()

    @staticmethod
    def _print_if_not_empty(label: str, value: str) -> str:
        return f", {label}='{value}'" if value else ""

    @staticmethod
    def _print_first_if_not_empty(label: str, value: str) -> str:
        return f"{label}='{value}'" if value else ""

    @staticmethod
    def _print_if_true(label: str, value: bool) -> str:
        return f", {label}=True" if value else ""


from __future__ import annotations

from collections import defaultdict
import itertools
from dataclasses import dataclass
from typing import  Iterable, Sequence as seq

from relationalai.semantics.metamodel import metamodel as mm
from relationalai.semantics.metamodel.builtins import builtins as b
from relationalai.semantics.metamodel.rewriter import NO_WALK, Walker
from relationalai.util.structures import OrderedSet
from .helpers import is_output_update, is_main_output, ContainerWalker


@dataclass(frozen=True)
class VarRef:
    node: mm.Container
    input: bool = False

    def is_input(self) -> bool:
        return self.input

    def is_output(self) -> bool:
        return not self.input


class Scope:
    def __init__(self):
        # a variable to all references to it
        self.var_refs: dict[mm.Var, OrderedSet[VarRef]] = {}
        # a container to its parent container
        self.parent: dict[mm.Container, mm.Container] = {}
        # the variable of a main output to the container that hosts it
        self.main_output: dict[mm.Var, mm.Container] = {}
        self.available_vars: dict[mm.Container, OrderedSet[mm.Var]] = defaultdict(OrderedSet[mm.Var])

    def is_ancestor(self, ancestor: mm.Container, descendant: mm.Container) -> bool:
        if ancestor == descendant:
            return True
        n = descendant
        while n in self.parent:
            if n == ancestor:
                return True
            n = self.parent[n]
        return n == ancestor

    def least_common_ancestor(self, c1: mm.Container, c2: mm.Container) -> mm.Container | None:
        ancestors1 = set()
        n = c1
        while n in self.parent:
            ancestors1.add(n)
            n = self.parent[n]
        ancestors1.add(n)
        n = c2
        while n not in ancestors1:
            n = self.parent[n]
        return n


class Hoister(ContainerWalker):
    def __init__(self):
        super().__init__()
        # the result of the whole analysis
        self.hoists: dict[mm.Container, OrderedSet[mm.Var]] = defaultdict(OrderedSet[mm.Var])
        self.container_vars: dict[mm.Container, set[mm.Var]] = defaultdict(set)
        self.allowed_vars: dict[mm.Container, set[mm.Var]] = {}
        # the current scope
        self.scope: Scope = Scope()
        # stack of containers being traversed
        self.stack: list[mm.Container] = []

    #------------------------------------------------------
    # Stack
    #------------------------------------------------------

    def top(self) -> mm.Container | None:
        if self.stack:
            return self.stack[-1]
        return None

    def grand_parent(self) -> mm.Container | None:
        if len(self.stack) > 1:
            return self.stack[-2]
        return None

    #------------------------------------------------------
    # Public API
    #------------------------------------------------------

    def analyze(self, model: mm.Model):
        self.clear()
        self(model.root)
        self.scope = Scope()

    def analyze_task(self, task: mm.Task):
        self.clear()
        self(task)
        self.scope = Scope()

    def clear(self):
        self.hoists.clear()
        self.scope = Scope()

    def hoisted(self, container: mm.Container) -> OrderedSet[mm.Var]:
        return self.hoists[container]

    #------------------------------------------------------
    # Walking
    #------------------------------------------------------

    def enter_container(self, container: mm.Container, children: seq[mm.Node]):
        # potentially reset scope (can we nest scopes?)
        if container.scope:
            self.scope = Scope()
        # compute parent relationship
        parent = self.top()
        if parent:
            self.scope.parent[container] = parent
        # push container onto stack
        self.stack.append(container)

    def exit_container(self, container: mm.Container, children: seq[mm.Node]):
        if isinstance(container, (mm.Match, mm.Union)):
            var_sets = [set(self.scope.available_vars[child]) for child in container.tasks if isinstance(child, mm.Container)]
            # allowed vars are those that appear in all branches
            allowed = set.intersection(*var_sets) if var_sets else set()
            parent_hoists = self.hoists[container]
            for child in container.tasks:
                if isinstance(child, mm.Container):
                    self.hoists[child] = parent_hoists
                    self.allowed_vars[child] = allowed
            self.allowed_vars[container] = allowed

        if container.scope:
            # when leaving the scope, compute hoisted vars
            for var, var_refs in self.scope.var_refs.items():
                # find all places where this var is used as input but there's no output for
                # the var in an ancestor container; this means we need to hoist the var from
                # some output
                requires_hoist: OrderedSet[VarRef] = OrderedSet()
                for ref in var_refs:
                    if ref.is_input() and not any(self.scope.is_ancestor(other.node, ref.node) for other in var_refs if other != ref and other.is_output()):
                        requires_hoist.add(ref)

                # for references that require a hoist, find some other output and hoist
                # that up until the least common ancestor of the two references
                for ref in requires_hoist:
                    for other in var_refs:
                        if other != ref and other.is_output():
                            # find least common ancestor of the two references
                            lca = self.scope.least_common_ancestor(ref.node, other.node)
                            assert(lca is not None)
                            if var not in self.container_vars[lca]:
                                self._hoist_until(var, other.node, lca)

        # pop container from stack
        self.stack.pop()

    def _hoist_until(self, var: mm.Var, from_container: mm.Container, to_container: mm.Container):
        """ Hoist the variable from `from_container` up to (but not including) `to_container`. """
        n = from_container
        hoisted = True
        while n != to_container:
            if isinstance(n, mm.Not):
                hoisted = False
                break
            if self.allowed_vars.get(n) is not None:
                allowed = self.allowed_vars[n]
                if var not in allowed:
                    hoisted = False
                    break
            self.hoists[n].add(var)
            n = self.scope.parent[n]
        if hoisted:
            self.container_vars[to_container].add(var)

    #------------------------------------------------------
    # Tasks that manipulate variables
    #------------------------------------------------------

    def _register_use(self, var: mm.Value, input: bool, container: mm.Container | None =None):
        if isinstance(var, mm.Var):
            parent = self.top() if container is None else container
            assert(parent is not None)

            if var not in self.scope.var_refs:
                self.scope.var_refs[var] = OrderedSet()
            self.scope.var_refs[var].add(VarRef(parent, input))
            self.scope.available_vars[parent].add(var)

    def lookup(self, l: mm.Lookup):
        for arg, field in zip(l.args, l.relation.fields):
            if not field.input or l.relation == b.core["="]:
                self._register_use(arg, input=False)
            else:
                self._register_use(arg, input=True)

    def update(self, u: mm.Update):
        if is_main_output(u):
            # just register the container for the main output, using the output variable as key
            self.scope.main_output[u.args[0]] = self.top() # type: ignore
        else:
            main_output_container = self.scope.main_output[u.args[0]] if is_output_update(u) else None # type: ignore
            for arg in u.args:
                self._register_use(arg, input=True)
                # if this is an output update, register that the container hosting the main
                # output needs this variable as input
                if main_output_container is not None:
                    self._register_use(arg, input=True, container=main_output_container)

    def aggregate(self, a: mm.Aggregate):
        self._register_use(a.projection, input=True)
        for g in a.group:
            self._register_use(g, input=True)
        for arg, field in zip(a.args, a.aggregation.fields):
            self._register_use(arg, input=field.input)

    def construct(self, c: mm.Construct):
        for v in c.values:
            self._register_use(v, input=True)
        self._register_use(c.id_var, input=False)



from v0.relationalai.environments.base import except_return
from v0.relationalai.environments.ipython import IPythonEnvironment
import sys

class ColabEnvironment(IPythonEnvironment):
    remote = True

    @classmethod
    def detect(cls):
        return "google.colab" in sys.modules

    @except_return(Exception, None)
    def active_cell_id(self):
        return self.ipython.get_parent()["metadata"]["colab"]["cell_id"] #type: ignore



from v0.relationalai.dsl import safe_symbol
from v0.relationalai.rel_utils import sanitize_identifier
from v0.relationalai.std import rel


def SFRawType(model, *, source):
    raw_relation_name = sanitize_identifier(source.lower())

    # We run the normal machinery to setup the source + retrieve metadata, but never actually consume the relations
    PyRelType = model.Type(f"__PyRel__{raw_relation_name}", source=source)

    sf_gnf_rel_name = sanitize_identifier(raw_relation_name + "_pyrel_id")

    # Generate the type from the raw table
    # @NOTE: The actual type isn't associated with the source, we handle mapping the raw CDC relations directly via
    #        zero-copy rules. This means the SFRawType entity ids are SF 160bit IDs instead of 128bit IDs, so
    #        intermixing will cause issues. Use with caution.
    # @NOTE: This work will be replaced by upstream work that does a similar thing but updates the CDC-created relations
    #        to be in the exact format we want.
    Typ = model.Type(f"__SF_RAW__{sf_gnf_rel_name}")
    relation = getattr(rel, sf_gnf_rel_name)  # unary relation of IDs
    with model.rule():
        Typ.add(relation())

    props = PyRelType.known_properties()

    # Generate the property mappings for the SF columns
    code = ""
    for prop in props:
        code += f"""
            def {sanitize_identifier(prop)}(row, value): {{
                {raw_relation_name}({safe_symbol(prop.upper())}, row, value)
            }}
        """

    model.install_raw(code)

    return Typ



import warnings

from v0.relationalai.semantics.lqp.ir import convert_transaction, validate_lqp

__all__ = ["convert_transaction", "validate_lqp"]

warnings.warn(
    "relationalai.early_access.lqp.ir is deprecated. "
    "Please migrate to relationalai.semantics.lqp.ir",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations
from typing import Optional, Any, Union

import re
import v0.relationalai.semantics as qb
from v0.relationalai.semantics.internal.internal import RelationshipFieldRef, Field
from v0.relationalai.early_access.dsl.core.utils import generate_stable_uuid
from v0.relationalai.early_access.dsl.orm.utils import generate_rai_way_name

##########
# Utility functions for parsing RelationshipReading patterns
#

def leading_text_of(reading: RelationshipReading):
    pattern = r"^([^{]*).*$"
    return re.findall(pattern, reading._madlib)[0].strip()

# Returns the dictionary of texts following a role (empty string if not present)
# indexed by the role position
#
def roles_data_from(reading: RelationshipReading):
    # This pattern capture the player name (with optional role name) and the following text for each
    # role in the madlib string
    # \{([^{}]*)\}   Player or role_name:Player
    # ([^{]*)       Following text
    pattern = r"\{([^{}]*)\}([^{]*)"

    matches = re.findall(pattern, reading._madlib)
    following_texts = dict()

    for i, (_, follows) in enumerate(matches):
        following_texts[i] = follows.strip()
    return following_texts

# Analyzes a substring of relationship-reading pattern text that contains
# no role occurrences and refines it into a triple of 3 optional strings:
#
#  - post: is a substring of text that post-binds to the role occurrence
#    that this text follows
#  - inter: is a substring that is not part of any post or pre bound text
#  - pre: is a substring that pre-binds the role occurrence that follows
#    this text sequence in the original reading pattern.
#
# Any of these sequences could be empty on return.
#
def refine_reading_text(text):
    found_prefix = False
    post = None
    inter = None
    pre = None

    inter_seq = []
    pre_seq = []

    tokens = text.split()
    for v in tokens:
        if v.startswith('-'):
            if found_prefix:
                raise Exception(f"Malformed reading text {text} has prefix text preceding postfix text")
            inter_seq.append(v[1:])
            post = " ".join(inter_seq)
            inter_seq = []
        elif v.endswith('-'):
            if found_prefix:
                raise Exception(f"Malformed reading text {text} has multple postfix text delimeters")
            found_prefix = True
            pre_seq = [ v[0:len(v)-1] ]
        else:
            if found_prefix:
                pre_seq.append(v)
            else:
                inter_seq.append(v)

    if len(pre_seq) > 0:
        pre = " ".join(pre_seq)
    if len(inter_seq) > 0:
        inter = " ".join(inter_seq)

    return (post, inter, pre)


class Relationship(qb.Relationship):

    def __init__(self, model, madlib:Any, short_name:str="", fields:Optional[list[Field]]=None):
        super().__init__(madlib, short_name=short_name, model=model.qb_model(), fields=fields)
        self._dsl_model = model
        self._rel_roles = {field.name: self.__getitem__(field.name) for field in self._fields}
        self._readings[0] = RelationshipReading(self._dsl_model, madlib, self, short_name)

    def __getitem__(self, arg:Union[str, int, qb.Concept]) -> Any:
        rel_field_ref = super().__getitem__(arg)
        field_name = rel_field_ref._field_ref._name
        if hasattr(self, "_rel_roles"):
            if field_name in self._rel_roles:
                return self._rel_roles[field_name]
            else:
                raise ValueError(f"{arg} is undefined for {self._name}")
        return Role._from_field(rel_field_ref)

    def _guid(self):
        return generate_stable_uuid(str(self._id))

    def alt(self, madlib:Any, short_name:str = "", reading:qb.RelationshipReading|None = None) -> qb.RelationshipReading:
        return super().alt(madlib, short_name=short_name,
                           reading=RelationshipReading(self._dsl_model, madlib, self, short_name))

    def _unary(self):
        return self._arity() == 1

    def _binary(self):
        return self._arity() == 2

    def _first(self):
        return self.__getitem__(0)

    def _roles(self):
        return [self._rel_roles[field.name] for field in self._fields]


class RelationshipReading(qb.RelationshipReading):

    def __init__(self, model, madlib:Any, alt_of:Relationship, short_name:str):
        super().__init__(madlib, alt_of, short_name, model=model.qb_model())
        self._dsl_model = model
        self._prepare()

    def __getitem__(self, arg:Union[str, int, qb.Concept]) -> Any:
        return Role._from_field(super().__getitem__(arg))

    def _guid(self):
        return generate_stable_uuid(str(self._id))

    def _unary(self):
        return self._arity() == 1

    def _binary(self):
        return self._arity() == 2

    def _first(self):
        return self.__getitem__(0)

    def _last(self):
        roles = self._roles()
        return roles[len(roles) - 1]

    def _roles(self):
        return [self._alt_of._rel_roles[field.name] for field in self._fields]

    def rai_way_name(self):
        return generate_rai_way_name(self)

    def _sample_fact(self) -> str:
        verb = []
        for idx in range(len(self._role_in_reading)):
            verb.append(str(self._role_in_reading[idx]))
            follows = self._follows[idx]
            if follows is not None:
                verb.append(follows)
        return ' '.join(verb)

    # Analyzes the reading pattern supplied with this reading to compute:
    #  - the leading text (or None)
    #  - the sequence of role occurrences in this reading order, and
    #  - the sequence of text fragments that follow each role occurrence
    #    in this reading order
    #
    def _prepare(self):

        self._follows: list[Optional[str]] = []
        self._role_in_reading: list[RoleOccurrence] = []

        (post, self._leading_text, pre) = refine_reading_text(leading_text_of(self))
        pre_seq = {0: pre}
        role_seq = self._roles()

        idx = 0
        dict = roles_data_from(self)
        for v in dict.values():
            (post, inter, pre) = refine_reading_text(v)
            self._role_in_reading.append(RoleOccurrence(role_seq[idx], pre_seq[idx], post))
            self._follows.append(inter)
            pre_seq[idx + 1] = pre
            idx += 1


class Role(RelationshipFieldRef):
    _sibling: Optional[Role] = None

    def __init__(self, parent:Any, part_of, pos):
        super().__init__(parent, part_of, pos)

    def _guid(self):
        return generate_stable_uuid(f"{self._field_ix}_{self._part_of()._guid()}")

    def player(self) -> qb.Concept:
        return self._concept

    def sibling(self):
        if self._relationship._arity() == 2 and not self._sibling:
            first_role = self._relationship[0]
            sibling = self._relationship[1] if self._id == first_role._id else first_role
            self._sibling = sibling
        return self._sibling
    
    def siblings(self):
        return [self._relationship[i] for i in range(self._relationship._arity()) if i != self._field_ix]

    def _part_of(self):
        return self._relationship

    @staticmethod
    def _from_field(field:RelationshipFieldRef):
        return Role(field._parent, field._relationship, field._field_ix)

    def __hash__(self):
        return hash(f"Role({self._guid()})")

    def __eq__(self, other):
        if not isinstance(other, Role):
            return False
        return self._guid() == other._guid() and self._part_of() == other._part_of()

# A RoleOccurrence objectifies the occurrence of a Role in some RelationshipReading
# to account for any pre- and/or post-bound text around the Role occurrence in
# that reading. For instance, in the RelationshipReading with this reading pattern:
#
#     "{Person} has personal- {Vehicle}"
#
# the RoleOccurrences for the roles verbalize as
#
#     "Person" and
#     "personal Vehicle"
#
class RoleOccurrence:

    def __init__(self, role: Role, pre: Optional[str], post: Optional[str]):
        self._prefix: Optional[str] = pre
        self._postfix: Optional[str] = post
        self._role: Role = role
        self._verbalizes_as: Optional[str] = None

    def __str__(self):
        if self._verbalizes_as is None:
            verb_seq = []
            if self._prefix is not None:
                verb_seq.append(self._prefix)
            verb_seq.append(str(self._role.player()))
            if self._postfix is not None:
                verb_seq.append(self._postfix)
            self._verbalizes_as = " ".join(verb_seq)
        return self._verbalizes_as



from v0.relationalai.semantics.metamodel import ir
from typing import Union

CompilableType = Union[
    # Subgrouping
    ir.Logical,
    ir.Union,

    # Formulas
    ir.Lookup,
    ir.Exists,
    ir.Construct,
    ir.Not,

    # Aggregations
    ir.Aggregate,

    # Rank
    ir.Rank,

    # Effects
    ir.Output,
    ir.Update,

    # Constraints
    ir.Require,
]

# Preconditions
def assert_valid_input(model: ir.Model) -> None:
    task = model.root

    assert isinstance(task, ir.Logical), f"expected root to be logical task, got {type(task)}"
    for subtask in task.body:
        _assert_valid_subtask(subtask)

def _assert_valid_subtask(task: ir.Task) -> None:
    # TODO: assert what subtasks should look like
    assert isinstance(task, ir.Logical), f"expected logical task, got {type(task)}"
    _assert_task_compilable(task)

def _assert_task_compilable(task: ir.Task) -> None:
    assert isinstance(task, CompilableType), f"expected task to be compilable, got {type(task)}"
    if isinstance(task, ir.Logical):
        for subtask in task.body:
            _assert_task_compilable(subtask)
    elif isinstance(task, ir.Union):
        for subtask in task.tasks:
            _assert_task_compilable(subtask)
    elif isinstance(task, ir.Update):
        assert_valid_update(task)
        effect = task.effect
        assert effect == ir.Effect.derive, "only derive supported at the moment"

def assert_valid_update(update: ir.Update) -> None:
    effect = update.effect
    assert effect == ir.Effect.derive, "only derive supported at the moment"



from v0.relationalai.semantics.metamodel import factory as f
from v0.relationalai.semantics.metamodel.util import FrozenOrderedSet
from v0.relationalai.semantics.metamodel import builtins

# Indicates a relation is short-lived, thus, backends should not optimize for incremental
# maintenance.
adhoc = f.relation("adhoc", [])
adhoc_annotation = f.annotation(adhoc, [])

# We only want to emit attributes for a known set of annotations.
annotations_to_emit = FrozenOrderedSet([
    adhoc.name,
    builtins.function.name,
    builtins.track.name,
    builtins.recursion_config.name,
])



from __future__ import annotations
import csv
from datetime import date, datetime
from typing import Any, cast

import numpy
import pandas

from v0.relationalai.clients.types import ImportSource, ImportSourceFile
from v0.relationalai.clients.client import ResourcesBase
from v0.relationalai.dsl import Graph, RelationRef, build, tag
from v0.relationalai.loaders.loader import TYPE_TO_REL_SCHEMA, Loader, compute_file_hash, compute_str_hash
from v0.relationalai.rel_utils import Char, assert_no_problems, emit_nested_relation
from v0.relationalai.loaders.types import Schema, Syntax
from v0.relationalai.metamodel import ActionType, Builtins, Var
from v0.relationalai.rel_emitter import sanitize
from v0.relationalai.std import rel


def df_type(df_type_name):
    if df_type_name == "bool":
        type = Builtins.Bool
    elif df_type_name == "int64":
        type = Builtins.Number
    elif df_type_name == "float64":
        type = Builtins.Number
    else:
        type = Builtins.String
    return type

class ExternalRow:
    def __init__(self, data, columns):
        self._data = data
        self._columns = columns

    def __getitem__(self, index):
        if isinstance(index, str):
            return getattr(self, index)
        return self._data[index]

    def __getattr__(self, name):
        if name in self._columns:
            return self._data[self._columns.index(name)]
        return object.__getattribute__(self, name)

# See https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
# for keyword arg docs.
def load_file(graph:Graph, csv_file, **kwargs):
    df = pandas.read_csv(csv_file, **kwargs)

    # create subqueries for each column that consist of a data object
    # and a bind for the column relation, also create a data ref
    # add a range for the id
    # when a dataref is used, it should add a get for the column relation
    # the rest just works?

    index = getattr(rel, sanitize(csv_file))
    tag(index, Builtins.Type)

    with graph.scope():
        index.add(rel.range(0, len(df), 1))

    id = Var(type=Builtins.Number)
    items = []
    for col in df.columns:
        sub = df[[col]]
        col_type = df_type(df[col].dtype)
        with graph.scope(dynamic=True):
            # By setting Builtins.RawData on the task, we're telling the denester to not
            # put a reference to the parent scope in this one
            graph._stack.active()._task.parents.append(Builtins.RawData)
            v1 = []
            v2 = []
            for (i, v) in sub.itertuples():
                if v is not None and v != "" and v is not False and (isinstance(v,str) or isinstance(v, date) or isinstance(v, datetime) or not numpy.isnan(v)):
                    v1.append(i)
                    v2.append(v)

            v1_var = Var(type=Builtins.Number)
            v2_var = Var(name=col, type=col_type)
            graph._action(
                build.relation_action(ActionType.Get, Builtins.RawData, [Var(value=v1), Var(value=v2), v1_var, v2_var])
            )
            temp = build.relation(sanitize(csv_file) + "_" + col, 2)
            temp.parents.append(Builtins.Property)
            graph._action(
                build.relation_action(ActionType.Bind, temp, [v1_var, v2_var])
            )
            usage_var = Var(name=col, type=col_type)
            items.append(RelationRef(graph, temp, [id, usage_var]))
    graph._action(
        build.relation_action(ActionType.Get, index._rel, [id])
    )
    return ExternalRow(items, [c for c in df.columns])

class CSVLoader(Loader):
    """Load CSV files and URLs in RAI DB using `load_csv`."""
    type = "csv"

    def load(self, provider: ResourcesBase, model: str, source: ImportSource, *, schema: Schema|None = None, syntax: Syntax|None = None, integration: dict|None = None):
        from railib import api
        assert isinstance(source, ImportSourceFile), "CSV Loader can only load from files and URLs, not {type(source).__name__}"
        schema = schema or {}
        syntax = syntax or {}

        id = compute_str_hash(source.raw_path)
        prefix = f"""def config[:path]: raw"{source.raw_path}" """
        inputs = None

        if not source.is_url():
            with open(source.raw_path, "r") as csv_file:
                data = csv_file.read()

            id = compute_file_hash(source.raw_path)
            prefix = "def config[:data]: data"
            inputs = {"data": data}

        relation = sanitize(source.name)

        for key in ["delim", "quotechar", "escapechar"]:
            if syntax.get(key, None) and not isinstance(syntax[key], Char):
                syntax[key] = Char(syntax[key])

        q = f"""
        declare {relation}
        def delete[:{relation}]: {relation}
        def delete[:__resource, k, "{relation}"]: __resource[k, "{relation}"]

        {prefix}
        {emit_nested_relation("config[:syntax, ", cast(dict, syntax), api._syntax_options)}
        {emit_nested_relation("config[:schema, ", {col: TYPE_TO_REL_SCHEMA.get(type, "string") for col, type in schema.items()})}
        {emit_nested_relation("config[:integration, ", integration) if integration else ""}
        {emit_nested_relation("insert[:__resource, ", {
        "id": (relation, id),
        "type": (relation, self.type),
        "name": (relation, source.name)
        })}
        def insert[:__resource, :schema]: ("{relation}", config[:schema])
        def insert[:{relation}]: load_csv[config]
        """
        q = "\n".join(line.strip() for line in q.splitlines())
        res = provider.exec_raw(model, provider.get_default_engine_name(), q, readonly=False, inputs=inputs)
        assert_no_problems(res)

    SAMPLE_LINES_COUNT = 100

    @classmethod
    def guess_syntax(cls, path: str) -> Syntax:
        sample_lines = []
        with open(path, "r") as file:
            try:
                for _ in range(cls.SAMPLE_LINES_COUNT):
                    sample_lines.append(next(file))
            except StopIteration:
                pass

        sample = "".join(sample_lines)
        sniffer = csv.Sniffer()
        dialect = sniffer.sniff(sample)

        # Unused but available: quoting, doublequote, lineterminator, skipinitialspace,
        return {
            "delim": dialect.delimiter,
            "quotechar": dialect.quotechar,
            "escapechar": dialect.escapechar,
            "header_row": 1 if sniffer.has_header(sample) else -1
        }

    @classmethod
    def guess_schema(cls, path: str, syntax: Syntax) -> tuple[Schema, pandas.DataFrame]:
        """Guess the schema by reading a chunk of the file."""
        # @TODO: pick an appropriate magic number of rows to consider for type information.

        pandas_syntax_args:dict[str, Any] = {
            "delimiter": syntax.get("delim", None),
            "quotechar": syntax.get("quotechar", None),
            "escapechar": syntax.get("escapechar", None),

            # @NOTE: For some reason the sniffer always wants \r\n even when its only \n, and pandas (like rel) requires single char line terminators
            # "lineterminator": dialect.lineterminator,
            # @NOTE: These can't be passed to load_csv so there's not much point helping pandas with them.
            # "doublequote": dialect.doublequote,
            # "quoting": dialect.quoting,
            # "skipinitialspace": dialect.skipinitialspace,
        }

        chunk = pandas.read_csv(path, nrows=cls.SAMPLE_LINES_COUNT, **pandas_syntax_args)
        chunk = chunk.rename(columns={k: k.strip() for k in chunk.columns})
        inferred_schema = Schema()
        for col in chunk.columns:
            inferred_schema[col] = df_type(chunk[col].dtype)

        return (inferred_schema, chunk)

CSVLoader.register_for_extensions(".csv", ".tsv")



# Azure import is dropped because we need to to do runtime import of azure module to support our package running on Snowflake Notebook
from . import config

# Lazy imports for client and local to avoid circular import issues
# These modules import from relationalai which imports from clients, creating a cycle
def __getattr__(name: str):
    if name == 'client':
        from . import client
        return client
    elif name == 'local':
        from . import local
        return local
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

# note: user must do `import relationalai.clients.azure` to get `azure` submodule
# note: snowflake module is now at relationalai.clients.resources.snowflake
# note: 'client' and 'local' are lazy-loaded via __getattr__, so they're not in __all__
__all__ = ['config']



"""
Pydantic models for Snowflake config.toml structure.

These models validate the structure of Snowflake config.toml files before
converting them to RAI Config format.
"""

from __future__ import annotations

from typing import Any
from pydantic import BaseModel, ConfigDict, Field, model_validator


def normalize_authenticator(authenticator: str | None) -> str:
    if not authenticator:
        return "username_password"

    auth = authenticator.lower()

    mapping = {
        "snowflake": "username_password",
        "username_password_mfa": "username_password_mfa",
        "externalbrowser": "externalbrowser",
        "snowflake_jwt": "jwt",
        "oauth": "oauth",
        "programmatic_access_token": "programmatic_access_token",
    }

    return mapping.get(auth, "username_password")


class SnowflakeConnection(BaseModel):
    model_config = ConfigDict(extra="allow")

    account: str
    user: str
    password: str | None = None
    warehouse: str
    role: str | None = None
    database: str | None = None
    schema_: str | None = Field(default=None, alias="schema")
    authenticator: str | None = None
    private_key_path: str | None = None
    private_key_passphrase: str | None = None
    token: str | None = None
    passcode: str | None = None
    
    
    @model_validator(mode='before')
    def check_required_fields(cls, values):
        required_fields = ['account', 'user', 'warehouse']
        missing_fields = [field for field in required_fields if field not in values or values[field] is None]
        if missing_fields:
            raise ValueError(f"Missing required fields in Snowflake connection: {', '.join(missing_fields)}")
        return values

    def convert(self) -> dict[str, Any]:
        connection_dict = self.model_dump(
            exclude_none=True,
            by_alias=True
        )

        connection_dict["type"] = "snowflake"
        connection_dict["authenticator"] = normalize_authenticator(self.authenticator)

        return connection_dict


class SnowflakeConfigFile(BaseModel):
    model_config = ConfigDict(extra="allow")

    default_connection_name: str | None = None
    connections: dict[str, SnowflakeConnection]
    
    @model_validator(mode='before')
    def check_connections(cls, values):
        if 'connections' not in values:
            raise ValueError("Snowflake config.toml is missing 'connections' section")
        return values

    @model_validator(mode='before')
    def check_default_connection(cls, values):
        if 'default_connection_name' not in values:
            raise ValueError("Snowflake config.toml is missing 'default_connection_name' field")
        return values



from __future__ import annotations

from ..frontend.base import Expression, Variable
from ..frontend.core import Integer

from . import StringValue, IntegerValue, numbers, _deprecated_library


# Coerce a number to Int64.
def int64(value: IntegerValue) -> Variable:
    _warning()
    return Integer(value)

# Coerce a number to Int128.
def int128(value: IntegerValue) -> Variable:
    _warning()
    return Integer(value)

def parse_int64(value: StringValue) -> Expression:
    _warning(f"numbers.parse_number({value}, , 0)")
    return numbers.parse_number(value, 19, 0)

def parse_int128(value: StringValue) -> Expression:
    _warning(f"numbers.parse_number({value}, 38, 0)")
    return numbers.parse_number(value, 38, 0)

# Alias parse_int128 to parse
def parse(value: StringValue) -> Expression:
    return parse_int128(value)


#--------------------------------------------------
# Implementation
#--------------------------------------------------

def _warning(msg: str|None = None):
    _deprecated_library("std.integers", "std.numbers", msg)



"""
RelationalAI Graph Library
"""

__version__ = "0.0.0"

import warnings
from v0.relationalai.docutils import ProductStage
from .core import Graph

# Mark this package's docstrings for inclusion
# in automatically generated web documentation.
__include_in_docs__ = True
__rai_product_stage__ = ProductStage.EARLY_ACCESS

# Warn on import that this package is at an early stage of development,
# intended for internal consumers only, and ask those internal consumers
# to contact the symbolic reasoning team such that we can track usage,
# get feedback, and help folks through breaking changes.
warnings.warn(
    (
        "\n\nThis library is still in early stages of development and is intended "
        "for internal use only. Among other considerations, interfaces will change, "
        "and performance is appropriate only for exploring small graphs. Please "
        "see this package's README for additional information.\n\n"
        "If you are an internal user seeing this, please also contact "
        "the symbolic reasoning team such that we can track usage, get "
        "feedback, and help you through breaking changes.\n"
    ),
    FutureWarning,
    stacklevel=2
)

# Finally make this package's core functionality publicly available.
__all__ = [
    "Graph",
]



from v0.relationalai.semantics import Model, Integer, define, select, count
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.path_algorithms.find_paths import find_shortest_paths


# First test with grid graph
model = Model("test_all", dry_run=False)

grid = Graph.construct_grid(model, 3)

path_num, k = Integer.ref(), Integer.ref()
v = grid.Node.ref()

n1_1 = grid.Node.new(row = 1, col = 1)
n1_2 = grid.Node.new(row = 1, col = 2)
n1_3 = grid.Node.new(row = 1, col = 3)
n2_1 = grid.Node.new(row = 2, col = 1)
n2_2 = grid.Node.new(row = 2, col = 2)
n2_3 = grid.Node.new(row = 2, col = 3)
n3_1 = grid.Node.new(row = 3, col = 1)
n3_2 = grid.Node.new(row = 3, col = 2)
n3_3 = grid.Node.new(row = 3, col = 3)

Source_1 = model.Concept("Source_1", extends=[grid.Node])
Target_1 = model.Concept("Target_1", extends=[grid.Node])

define(Source_1(n1_1))
define(Target_1(n3_3))

paths_1 = find_shortest_paths(grid, Source_1, Target_1, num_paths=1)
iter_path_1 = select(k, v.row, v.col).where(paths_1(path_num, k, v)).to_df()
tuple_path_1 = tuple(sorted([row for row in iter_path_1.itertuples(index = False, name = None)]))

expected_path_1 = (
    (0, 1, 1), (1, 1, 2), (2, 1, 3), (3, 2, 3), (4, 3, 3)
)

expected_path_2 = (
    (0, 1, 1), (1, 1, 2), (2, 2, 2), (3, 2, 3), (4, 3, 3)
)

expected_path_3 = (
    (0, 1, 1), (1, 1, 2), (2, 2, 2), (3, 3, 2), (4, 3, 3)
)

expected_path_4 = (
    (0, 1, 1), (1, 2, 1), (2, 2, 2), (3, 2, 3), (4, 3, 3)
)

expected_path_5 = (
    (0, 1, 1), (1, 2, 1), (2, 2, 2), (3, 3, 2), (4, 3, 3)
)

expected_path_6 = (
    (0, 1, 1), (1, 2, 1), (2, 3, 1), (3, 3, 2), (4, 3, 3)
)

assert (
    tuple_path_1 == expected_path_1 or
    tuple_path_1 == expected_path_2 or
    tuple_path_1 == expected_path_3 or
    tuple_path_1 == expected_path_4 or
    tuple_path_1 == expected_path_5 or
    tuple_path_1 == expected_path_6
)



# Second test with grid graph
paths_2 = find_shortest_paths(grid, Source_1, Target_1, num_paths=3)

assert select(count(path_num)).where(paths_2(path_num, k, v)).to_df().iloc[0,0] == 3



# Third test with grid graph
paths_3 = find_shortest_paths(grid, Source_1, Target_1, num_paths=6)
iter_path_3 = select(path_num, k, v.row, v.col).where(paths_3(path_num, k, v)).to_df()
list_path_3 = sorted([row for row in iter_path_3.itertuples(index = False, name = None)])
set_path_3 = set()
for i in range(6):
    set_path_3.add(tuple((tup[1], tup[2], tup[3]) for tup in list_path_3[5*i:5*(i+1)]))

assert (
    set_path_3 == {
        expected_path_1, expected_path_2, expected_path_3, expected_path_4, expected_path_5, expected_path_6
    }
)


# Fourth test with grid graph
paths_4 = find_shortest_paths(grid, Source_1, Target_1)
iter_path_4 = select(path_num, k, v.row, v.col).where(paths_4(path_num, k, v)).to_df()
list_path_4 = sorted([row for row in iter_path_4.itertuples(index = False, name = None)])
set_path_4 = set()
for i in range(6):
    set_path_4.add(tuple((tup[1], tup[2], tup[3]) for tup in list_path_4[5*i:5*(i+1)]))

assert (
    len(list_path_4) == 30 and
    set_path_4 == {
        expected_path_1, expected_path_2, expected_path_3, expected_path_4, expected_path_5, expected_path_6
    }
)



# For builder components.
from v0.relationalai.semantics import Integer, define, not_, count
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.utilities.iterators import setup_iteration


def two_balls_upto(g:Graph, Source, Target, max_length=None):
    edge = g.Edge
    Node = g.Node

    source_ball = g.model.Relationship(f"source_ball2 {{Integer}} {{{Node}}}")
    target_ball = g.model.Relationship(f"target_ball2 {{Integer}} {{{Node}}}")
    source_condition = g.model.Relationship("source_condition1 {Integer}")
    target_condition = g.model.Relationship("target_condition1 {Integer}")
    source_surface = g.model.Relationship("source_surface1 {Integer}")
    target_surface = g.model.Relationship("target_surface1 {Integer}")

    if max_length is None:
        radius_S = setup_iteration(g.model, source_condition, 0, 1000000)
        radius_T = setup_iteration(g.model, target_condition, 0, 1000000)
    else:
        half_radius = max_length//2
        radius_S = setup_iteration(g.model, source_condition, 0, half_radius)
        radius_T = setup_iteration(g.model, target_condition, 0, max_length - half_radius)

    u, v, w, src, tgt = Node.ref(), Node.ref(), Node.ref(), Node.ref(), Node.ref()
    m, n, k, l1, l2 = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()

    # Ball around src contains src at distance 0:
    define(source_ball(0, src)).where(Source(src))

    # Recursive case:
    define(source_ball(k, u)).where(
        radius_S(k),
        source_ball(k - 1, v),
        edge(v, u),
        not_(
            source_ball(n, u),
            n < k
        )
    )

    define(source_ball(k, u)).where(
        source_ball(k, u)
    )

    # Ball around dst contains dst at distance 0:
    define(target_ball(0, tgt)).where(Target(tgt))

    # Recursive case:
    define(target_ball(k, u)).where(
        radius_T(k),
        target_ball(k - 1, v),
        edge(u, v),
        not_(
            target_ball(n, u),
            n < k
        )
    )

    define(target_ball(k, u)).where(
        target_ball(k, u)
    )

    define(source_condition(n)).where(
        radius_S(n),
        source_ball(n, u),
        radius_T(m),
        target_ball(m, v),
        not_(
            source_ball(n, w),
            target_ball(m, w)
        ),
        source_surface(l1),
        target_surface(l2),
        l1 < l2
    )

    define(target_condition(m)).where(
        radius_S(n),
        source_ball(n, u),
        radius_T(m),
        target_ball(m, v),
        not_(
            source_ball(n, w),
            target_ball(m, w)
        ),
        source_surface(l1),
        target_surface(l2),
        l1 >= l2
    )

    define(target_surface(count(u))).where(
        radius_T(n),
        target_ball(n, u)
    )

    define(source_surface(count(u))).where(
        radius_S(n),
        source_ball(n, u)
    )

    return source_ball, target_ball


import base64
import hashlib
import jwt
from datetime import timedelta, timezone, datetime
from typing import Union, Optional, TYPE_CHECKING

# warehouse-based snowflake notebooks currently don't have hazmat
crypto_disabled = False
try:
    from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat, load_pem_private_key
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.asymmetric import rsa, ec, ed25519, ed448
except ModuleNotFoundError:
    crypto_disabled = True
    # Define fallback types/functions when cryptography is not available
    Encoding = PublicFormat = load_pem_private_key = default_backend = rsa = ec = ed25519 = ed448 = None

CLAIM_ISSUER = "iss"
CLAIM_EXPIRY = "exp"
CLAIM_ISSUED_AT = "iat"
CLAIM_SUBJECT = "sub"

# AllowedPrivateKeyTypes is defined to allow for proper typing.
# We need to differentiate between runtime and type-checking contexts, as Python 3.9 can
# not check instance types against Unions.
if TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric import rsa, ec, ed25519, ed448
    AllowedPrivateKeyTypes = Union[
        rsa.RSAPrivateKey,
        ec.EllipticCurvePrivateKey,
        ed25519.Ed25519PrivateKey,
        ed448.Ed448PrivateKey,
    ]
else:
    if not crypto_disabled:
        AllowedPrivateKeyTypes = (
            rsa.RSAPrivateKey,
            ec.EllipticCurvePrivateKey,
            ed25519.Ed25519PrivateKey,
            ed448.Ed448PrivateKey,
        )
    else:
        AllowedPrivateKeyTypes = tuple()

class JWTGenerator:
    """
    Manages JWT creation, signing, and automatic renewal with configurable lifetimes.
    """
    
    # Token configuration constants
    DEFAULT_TOKEN_LIFETIME = timedelta(minutes=59)
    DEFAULT_RENEWAL_THRESHOLD = timedelta(minutes=54)
    SIGNING_ALGORITHM = "RS256"

    def __init__(
        self, 
        account: str, 
        username: str, 
        private_key_path: str, 
        key_passphrase: Optional[str] = None,
        token_lifetime: timedelta = DEFAULT_TOKEN_LIFETIME, 
        renewal_threshold: timedelta = DEFAULT_RENEWAL_THRESHOLD
    ):
        """
        Initialize JWT authentication generator.
        
        Args:
            account: Account identifier for authentication
            username: User identifier
            private_key_path: Filesystem path to private key file
            key_passphrase: Optional passphrase for encrypted private keys
            token_lifetime: Duration for which tokens remain valid
            renewal_threshold: Time before expiry to trigger renewal
        """
        
        self.account = account.upper()
        self.username = username.upper()
        self.fully_qualified_user = f"{self.account}.{self.username}"

        self.token_lifetime = token_lifetime
        self.renewal_threshold = renewal_threshold
        self._key_file_path = private_key_path
        self._next_renewal_time = datetime.min
        self._cached_token: Optional[str] = None

        self._private_key = self._load_private_key_from_file(key_passphrase)

    def _load_private_key_from_file(self, passphrase: Optional[str]) -> AllowedPrivateKeyTypes:
        if crypto_disabled or load_pem_private_key is None or default_backend is None:
            raise Exception("Cryptography library is not available. please install the 'cryptography' package to enable JWT support.")
        with open(self._key_file_path, "rb") as key_file:
            key_data = key_file.read()

            try:
                private_key = load_pem_private_key(key_data, None, default_backend())
            except TypeError:
                if passphrase is None:
                    raise ValueError("Private key requires passphrase but none provided")
                encoded_passphrase = passphrase.encode("utf-8")
                private_key = load_pem_private_key(key_data, encoded_passphrase, default_backend())

        if not isinstance(private_key, AllowedPrivateKeyTypes):
            raise TypeError(f"Unsupported private key type: {type(private_key)}")

        return private_key

    def get_or_generate_token(self) -> str:
        """
        Generate or return cached JWT access token.
        Automatically renews token when renewal threshold is reached.
        
        Returns:
            Valid JWT access token
        """
        if crypto_disabled:
            raise Exception("Cryptography library is not available. please install the 'cryptography' package to enable JWT support.")

        current_time = datetime.now(timezone.utc)

        # we have a valid cached token
        if self._cached_token and self._next_renewal_time > current_time:
            return self._cached_token

        # Generate public key fingerprint for issuer field
        key_fingerprint = self._compute_public_key_fingerprint()

        # Construct JWT payload
        token_payload = {
            CLAIM_ISSUER: f"{self.fully_qualified_user}.{key_fingerprint}",
            CLAIM_SUBJECT: self.fully_qualified_user,
            CLAIM_ISSUED_AT: current_time,
            CLAIM_EXPIRY: current_time + self.token_lifetime
        }

        # Sign and encode the token
        signed_token = jwt.encode(
            token_payload, 
            key=self._private_key, 
            algorithm=self.SIGNING_ALGORITHM
        )
        
        # Handle byte string response from older PyJWT versions
        if isinstance(signed_token, bytes):
            signed_token = signed_token.decode('utf-8')
            
        self._cached_token = signed_token
        self._next_renewal_time = current_time + self.renewal_threshold

        return self._cached_token

    def _compute_public_key_fingerprint(self) -> str:
        """
        Compute SHA256 fingerprint of the public key.
        
        Returns:
            Base64-encoded SHA256 fingerprint with 'SHA256:' prefix
        """
        if crypto_disabled or Encoding is None or PublicFormat is None:
            raise Exception("Cryptography library is not available. please install the 'cryptography' package to enable JWT support.")
        
        # Extract public key in DER format
        public_key_der = self._private_key.public_key().public_bytes(
            Encoding.DER, 
            PublicFormat.SubjectPublicKeyInfo
        )

        # Compute SHA256 hash
        hasher = hashlib.sha256()
        hasher.update(public_key_der)
        digest = hasher.digest()

        # Encode as base64 with prefix
        fingerprint = 'SHA256:' + base64.b64encode(digest).decode('utf-8')

        return fingerprint


"""
Direct Access Resources - Resources class for Direct Service Access.
This class overrides methods to use direct HTTP requests instead of Snowflake service functions.
"""
from __future__ import annotations
from typing import Any, Dict, List, Optional, Union
import requests

from .... import debugging
from ....tools.constants import USE_GRAPH_INDEX, DEFAULT_QUERY_TIMEOUT_MINS, Generation
from ....environments import runtime_env, SnowbookEnvironment
from ...config import Config, ConfigStore, ENDPOINT_FILE
from ...direct_access_client import DirectAccessClient
from ...types import EngineState
from ...util import get_pyrel_version, poll_with_specified_overhead, safe_json_loads, ms_to_timestamp
from ....errors import ResponseStatusException, QueryTimeoutExceededException
from snowflake.snowpark import Session

# Import UseIndexResources to enable use_index functionality with direct access
from .use_index_resources import UseIndexResources

# Import helper functions from util
from .util import is_engine_issue as _is_engine_issue, is_database_issue as _is_database_issue, collect_error_messages

from .use_index_poller import DirectUseIndexPoller
from typing import Iterable

# Constants
TXN_ABORT_REASON_TIMEOUT = "transaction timeout"


class DirectAccessResources(UseIndexResources):
    """
    Resources class for Direct Service Access avoiding Snowflake service functions.
    Uses HTTP requests instead of Snowflake SQL for execution.
    """
    def __init__(
        self,
        profile: Union[str, None] = None,
        config: Union[Config, None] = None,
        connection: Union[Session, None] = None,
        dry_run: bool = False,
        reset_session: bool = False,
        generation: Optional[Generation] = None,
        language: str = "rel",
    ):
        super().__init__(
            generation=generation,
            profile=profile,
            config=config,
            connection=connection,
            reset_session=reset_session,
            dry_run=dry_run,
            language=language,
        )
        self._endpoint_info = ConfigStore(ENDPOINT_FILE)
        self._service_endpoint = ""
        self._direct_access_client = None
        # database and language are already set by UseIndexResources.__init__

    @property
    def service_endpoint(self) -> str:
        return self._retrieve_service_endpoint()

    def _retrieve_service_endpoint(self, enforce_update=False) -> str:
        account = self.config.get("account")
        app_name = self.config.get("rai_app_name")
        service_endpoint_key = f"{account}.{app_name}.service_endpoint"
        if self._service_endpoint and not enforce_update:
            return self._service_endpoint
        if self._endpoint_info.get(service_endpoint_key, "") and not enforce_update:
            self._service_endpoint = str(self._endpoint_info.get(service_endpoint_key, ""))
            return self._service_endpoint

        is_snowflake_notebook = isinstance(runtime_env, SnowbookEnvironment)
        query = f"CALL {self.get_app_name()}.app.service_endpoint({not is_snowflake_notebook});"
        result = self._exec(query)
        assert result, f"Could not retrieve service endpoint for {self.get_app_name()}"
        if is_snowflake_notebook:
            self._service_endpoint = f"http://{result[0]['SERVICE_ENDPOINT']}"
        else:
            self._service_endpoint = f"https://{result[0]['SERVICE_ENDPOINT']}"

        self._endpoint_info.set(service_endpoint_key, self._service_endpoint)
        # save the endpoint to `ENDPOINT_FILE` to avoid calling the endpoint with every
        # pyrel execution
        try:
            self._endpoint_info.save()
        except Exception:
            print("Failed to persist endpoints to file. This might slow down future executions.")

        return self._service_endpoint

    @property
    def direct_access_client(self) -> DirectAccessClient:
        if self._direct_access_client:
            return self._direct_access_client
        try:
            service_endpoint = self.service_endpoint
            self._direct_access_client = DirectAccessClient(
                self.config, self.token_handler, service_endpoint, self.generation,
            )
        except Exception as e:
            raise e
        return self._direct_access_client

    def request(
        self,
        endpoint: str,
        payload: Dict[str, Any] | None = None,
        headers: Dict[str, str] | None = None,
        path_params: Dict[str, str] | None = None,
        query_params: Dict[str, str] | None = None,
        skip_auto_create: bool = False,
        skip_engine_db_error_retry: bool = False,
    ) -> requests.Response:
        with debugging.span("direct_access_request"):
            def _send_request():
                return self.direct_access_client.request(
                    endpoint=endpoint,
                    payload=payload,
                    headers=headers,
                    path_params=path_params,
                    query_params=query_params,
                )
            try:
                response = _send_request()
                if response.status_code != 200:
                    # For 404 responses with skip_auto_create=True, return immediately to let caller handle it
                    # (e.g., get_engine needs to check 404 and return None for auto_create_engine)
                    # For skip_auto_create=False, continue to auto-creation logic below
                    if response.status_code == 404 and skip_auto_create:
                        return response

                    try:
                        message = response.json().get("message", "")
                    except requests.exceptions.JSONDecodeError:
                        # Can't parse JSON response. For skip_auto_create=True (e.g., get_engine),
                        # this should have been caught by the 404 check above, so this is an error.
                        # For skip_auto_create=False, we explicitly check status_code below,
                        # so we don't need to parse the message.
                        if skip_auto_create:
                            raise ResponseStatusException(
                                f"Failed to parse error response from endpoint {endpoint}.", response
                            )
                        message = ""  # Not used when we check status_code directly

                    # fix engine on engine error and retry
                    # Skip setting up GI if skip_auto_create is True to avoid recursion or skip_engine_db_error_retry is true to let _exec_async_v2 perform the retry with the correct headers.
                    if ((_is_engine_issue(message) and not skip_auto_create) or _is_database_issue(message)) and not skip_engine_db_error_retry:
                        engine_name = payload.get("caller_engine_name", "") if payload else ""
                        engine_name = engine_name or self.get_default_engine_name()
                        engine_size = self.config.get_default_engine_size()
                        # Use the mixin's _poll_use_index method
                        self._poll_use_index(
                            app_name=self.get_app_name(),
                            sources=self.sources,
                            model=self.database,
                            engine_name=engine_name,
                            engine_size=engine_size,
                            headers=headers,
                        )
                        response = _send_request()
            except requests.exceptions.ConnectionError as e:
                messages = collect_error_messages(e)
                if any("nameresolutionerror" in msg for msg in messages):
                    # when we can not resolve the service endpoint, we assume it is outdated
                    # hence, we try to retrieve it again and query again.
                    self.direct_access_client.service_endpoint = self._retrieve_service_endpoint(
                        enforce_update=True,
                    )
                    return _send_request()
                # raise in all other cases
                raise e
            return response

    def _txn_request_with_gi_retry(
        self,
        payload: Dict,
        headers: Dict[str, str],
        query_params: Dict,
        engine: Union[str, None],
    ):
        """Make request with graph index retry logic.

        Attempts request with gi_setup_skipped=True first. If an engine or database
        issue occurs, polls use_index and retries with gi_setup_skipped=False.
        """
        response = self.request(
            "create_txn", payload=payload, headers=headers, query_params=query_params, skip_auto_create=True, skip_engine_db_error_retry=True
        )

        if response.status_code != 200:
            try:
                message = response.json().get("message", "")
            except requests.exceptions.JSONDecodeError:
                message = ""

            if _is_engine_issue(message) or _is_database_issue(message):
                engine_name = engine or self.get_default_engine_name()
                engine_size = self.config.get_default_engine_size()
                # Use the mixin's _poll_use_index method
                self._poll_use_index(
                    app_name=self.get_app_name(),
                    sources=self.sources,
                    model=self.database,
                    engine_name=engine_name,
                    engine_size=engine_size,
                    headers=headers,
                )
                headers['gi_setup_skipped'] = 'False'
                response = self.request(
                    "create_txn", payload=payload, headers=headers, query_params=query_params, skip_auto_create=True, skip_engine_db_error_retry=True
                )
            else:
                raise ResponseStatusException("Failed to create transaction.", response)

        return response

    def _exec_async_v2(
        self,
        database: str,
        engine: Union[str, None],
        raw_code: str,
        inputs: Dict | None = None,
        readonly=True,
        nowait_durable=False,
        headers: Dict[str, str] | None = None,
        bypass_index=False,
        language: str = "rel",
        query_timeout_mins: int | None = None,
        gi_setup_skipped: bool = False,
    ):

        with debugging.span("transaction") as txn_span:
            with debugging.span("create_v2") as create_span:

                use_graph_index = self.config.get("use_graph_index", USE_GRAPH_INDEX)

                payload = {
                    "dbname": database,
                    "engine_name": engine,
                    "query": raw_code,
                    "v1_inputs": inputs,
                    "nowait_durable": nowait_durable,
                    "readonly": readonly,
                    "language": language,
                }
                if query_timeout_mins is None and (timeout_value := self.config.get("query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS)) is not None:
                    query_timeout_mins = int(timeout_value)
                if query_timeout_mins is not None:
                    payload["timeout_mins"] = query_timeout_mins
                query_params={"use_graph_index": str(use_graph_index and not bypass_index)}

                # Add gi_setup_skipped to headers
                if headers is None:
                    headers = {}
                headers["gi_setup_skipped"] = str(gi_setup_skipped)
                headers['pyrel_program_id'] = debugging.get_program_span_id() or ""

                response = self._txn_request_with_gi_retry(
                    payload, headers, query_params, engine
                )

                artifact_info = {}
                response_content = response.json()

                txn_id = response_content["transaction"]['id']
                state = response_content["transaction"]['state']

                txn_span["txn_id"] = txn_id
                create_span["txn_id"] = txn_id
                debugging.event("transaction_created", txn_span, txn_id=txn_id)

            # fast path: transaction already finished
            if state in ["COMPLETED", "ABORTED"]:
                if txn_id in self._pending_transactions:
                    self._pending_transactions.remove(txn_id)

                # Process rows to get the rest of the artifacts
                for result in response_content.get("results", []):
                    filename = result['filename']
                    # making keys uppercase to match the old behavior
                    artifact_info[filename] = {k.upper(): v for k, v in result.items()}

            # Slow path: transaction not done yet; start polling
            else:
                self._pending_transactions.append(txn_id)
                with debugging.span("wait", txn_id=txn_id):
                    poll_with_specified_overhead(
                        lambda: self._check_exec_async_status(txn_id, headers=headers), 0.1
                    )
                artifact_info = self._list_exec_async_artifacts(txn_id, headers=headers)

            with debugging.span("fetch"):
                return self._download_results(artifact_info, txn_id, state)

    def _prepare_index(
        self,
        model: str,
        engine_name: str,
        engine_size: str = "",
        language: str = "rel",
        rai_relations: List[str] | None = None,
        pyrel_program_id: str | None  = None,
        skip_pull_relations: bool = False,
        headers: Dict | None = None,
    ):
        """
        Prepare the index for the given engine and model.
        """
        with debugging.span("prepare_index"):
            if headers is None:
                headers = {}

            payload = {
                "model_name": model,
                "caller_engine_name": engine_name,
                "language": language,
                "pyrel_program_id": pyrel_program_id,
                "skip_pull_relations": skip_pull_relations,
                "rai_relations": rai_relations or [],
                "user_agent": get_pyrel_version(self.generation),
            }
            # Only include engine_size if it has a non-empty string value
            if engine_size and engine_size.strip():
                payload["caller_engine_size"] = engine_size

            response = self.request(
                "prepare_index", payload=payload, headers=headers
            )

            if response.status_code != 200:
                raise ResponseStatusException("Failed to prepare index.", response)

            return response.json()

    def _check_exec_async_status(self, txn_id: str, headers: Dict[str, str] | None = None) -> bool:
        """Check whether the given transaction has completed."""

        with debugging.span("check_status"):
            response = self.request(
                "get_txn",
                headers=headers,
                path_params={"txn_id": txn_id},
            )
            assert response, f"No results from get_transaction('{txn_id}')"

        response_content = response.json()
        transaction = response_content["transaction"]
        status: str = transaction['state']

        # remove the transaction from the pending list if it's completed or aborted
        if status in ["COMPLETED", "ABORTED"]:
            if txn_id in self._pending_transactions:
                self._pending_transactions.remove(txn_id)

        if status == "ABORTED" and transaction.get("abort_reason", "") == TXN_ABORT_REASON_TIMEOUT:
            config_file_path = getattr(self.config, 'file_path', None)
            timeout_ms = int(transaction.get("timeout_ms", 0))
            timeout_mins = timeout_ms // 60000 if timeout_ms > 0 else int(self.config.get("query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS) or DEFAULT_QUERY_TIMEOUT_MINS)
            raise QueryTimeoutExceededException(
                timeout_mins=timeout_mins,
                query_id=txn_id,
                config_file_path=config_file_path,
            )

        # @TODO: Find some way to tunnel the ABORT_REASON out. Azure doesn't have this, but it's handy
        return status == "COMPLETED" or status == "ABORTED"

    def _list_exec_async_artifacts(self, txn_id: str, headers: Dict[str, str] | None = None) -> Dict[str, Dict]:
        """Grab the list of artifacts produced in the transaction and the URLs to retrieve their contents."""
        with debugging.span("list_results"):
            response = self.request(
                "get_txn_artifacts",
                headers=headers,
                path_params={"txn_id": txn_id},
            )
            assert response, f"No results from get_transaction_artifacts('{txn_id}')"
            artifact_info = {}
            for result in response.json()["results"]:
                filename = result['filename']
                # making keys uppercase to match the old behavior
                artifact_info[filename] = {k.upper(): v for k, v in result.items()}
            return artifact_info

    def get_transaction_problems(self, txn_id: str) -> List[Dict[str, Any]]:
        with debugging.span("get_transaction_problems"):
            response = self.request(
                "get_txn_problems",
                path_params={"txn_id": txn_id},
            )
            response_content = response.json()
            if not response_content:
                return []
            return response_content.get("problems", [])

    def get_transaction_events(self, transaction_id: str, continuation_token: str = ''):
        response = self.request(
            "get_txn_events",
            path_params={"txn_id": transaction_id, "stream_name": "profiler"},
            query_params={"continuation_token": continuation_token},
        )
        response_content = response.json()
        if not response_content:
            return {
                "events": [],
                "continuation_token": None
            }
        return response_content

    #--------------------------------------------------
    # Databases
    #--------------------------------------------------

    def get_installed_packages(self, database: str) -> Union[Dict, None]:
        use_graph_index = self.config.get("use_graph_index", USE_GRAPH_INDEX)
        if use_graph_index:
            response = self.request(
                "get_model_package_versions",
                payload={"model_name": database},
            )
        else:
            response = self.request(
                "get_package_versions",
                path_params={"db_name": database},
            )
        if response.status_code == 404 and response.json().get("message", "") == "database not found":
            return None
        if response.status_code != 200:
            raise ResponseStatusException(
                f"Failed to retrieve package versions for {database}.", response
            )

        content = response.json()
        if not content:
            return None

        return safe_json_loads(content["package_versions"])

    def get_database(self, database: str):
        with debugging.span("get_database", dbname=database):
            if not database:
                raise ValueError("Database name must be provided to get database.")
            response = self.request(
                "get_db",
                path_params={},
                query_params={"name": database},
            )
            if response.status_code != 200:
                raise ResponseStatusException(f"Failed to get db. db:{database}", response)

            response_content = response.json()

            if (response_content.get("databases") and len(response_content["databases"]) == 1):
                db = response_content["databases"][0]
                return {
                    "id": db["id"],
                    "name": db["name"],
                    "created_by": db.get("created_by"),
                    "created_on": ms_to_timestamp(db.get("created_on")),
                    "deleted_by": db.get("deleted_by"),
                    "deleted_on": ms_to_timestamp(db.get("deleted_on")),
                    "state": db["state"],
                }
            else:
                return None

    def create_graph(self, name: str):
        with debugging.span("create_model", dbname=name):
            return self._create_database(name,"")

    def delete_graph(self, name:str, force=False, language: str = "rel"):
        prop_hdrs = debugging.gen_current_propagation_headers()
        if self.config.get("use_graph_index", USE_GRAPH_INDEX):
            keep_database = not force and self.config.get("reuse_model", True)
            with debugging.span("release_index", name=name, keep_database=keep_database, language=language):
                response = self.request(
                    "release_index",
                    payload={
                        "model_name": name,
                        "keep_database": keep_database,
                        "language": language,
                        "user_agent": get_pyrel_version(self.generation),
                    },
                    headers=prop_hdrs,
                )
                if (
                    response.status_code != 200
                    and not (
                        response.status_code == 404
                        and "database not found" in response.json().get("message", "")
                    )
                ):
                    raise ResponseStatusException(f"Failed to release index. Model: {name} ", response)
        else:
            with debugging.span("delete_model", name=name):
                self._delete_database(name, headers=prop_hdrs)

    def clone_graph(self, target_name:str, source_name:str, nowait_durable=True, force=False):
        if force and self.get_graph(target_name):
            self.delete_graph(target_name)
        with debugging.span("clone_model", target_name=target_name, source_name=source_name):
            return self._create_database(target_name,source_name)

    def _delete_database(self, name:str, headers:Dict={}):
        with debugging.span("_delete_database", dbname=name):
            response = self.request(
                "delete_db",
                path_params={"db_name": name},
                query_params={},
                headers=headers,
            )
            if response.status_code != 200:
                raise ResponseStatusException(f"Failed to delete db. db:{name} ", response)

    def _create_database(self, name:str, source_name:str):
        with debugging.span("_create_database", dbname=name):
            payload = {
                "name": name,
                "source_name": source_name,
            }
            response = self.request(
                "create_db", payload=payload, headers={}, query_params={},
            )
            if response.status_code != 200:
                raise ResponseStatusException(f"Failed to create db. db:{name}", response)

    #--------------------------------------------------
    # Engines
    #--------------------------------------------------

    def list_engines(self, state: str | None = None):
        response = self.request("list_engines")
        if response.status_code != 200:
            raise ResponseStatusException(
                "Failed to retrieve engines.", response
            )
        response_content = response.json()
        if not response_content:
            return []
        engines = [
            {
                "name": engine["name"],
                "id": engine["id"],
                "size": engine["size"],
                "state": engine["status"], # callers are expecting 'state'
                "created_by": engine["created_by"],
                "created_on": engine["created_on"],
                "updated_on": engine["updated_on"],
            }
            for engine in response_content.get("engines", [])
            if state is None or engine.get("status") == state
        ]
        return sorted(engines, key=lambda x: x["name"])

    def get_engine(self, name: str):
        response = self.request("get_engine", path_params={"engine_name": name, "engine_type": "logic"}, skip_auto_create=True)
        if response.status_code == 404: # engine not found return 404
            return None
        elif response.status_code != 200:
            raise ResponseStatusException(
                f"Failed to retrieve engine {name}.", response
            )
        engine = response.json()
        if not engine:
            return None
        engine_state: EngineState = {
            "name": engine["name"],
            "id": engine["id"],
            "size": engine["size"],
            "state": engine["status"], # callers are expecting 'state'
            "created_by": engine["created_by"],
            "created_on": engine["created_on"],
            "updated_on": engine["updated_on"],
            "version": engine["version"],
            "auto_suspend": engine["auto_suspend_mins"],
            "suspends_at": engine["suspends_at"],
        }
        return engine_state

    def _create_engine(
            self,
            name: str,
            size: str | None = None,
            auto_suspend_mins: int | None = None,
            is_async: bool = False,
            headers: Dict[str, str] | None = None
        ):
        # only async engine creation supported via direct access
        if not is_async:
            return super()._create_engine(name, size, auto_suspend_mins, is_async, headers=headers)
        payload:Dict[str, Any] = {
            "name": name,
        }
        if auto_suspend_mins is not None:
            payload["auto_suspend_mins"] = auto_suspend_mins
        if size is not None:
            payload["size"] = size
        response = self.request(
            "create_engine",
            payload=payload,
            path_params={"engine_type": "logic"},
            headers=headers,
            skip_auto_create=True,
        )
        if response.status_code != 200:
            raise ResponseStatusException(
                f"Failed to create engine {name} with size {size}.", response
            )

    def delete_engine(self, name:str, force:bool = False, headers={}):
        response = self.request(
            "delete_engine",
            path_params={"engine_name": name, "engine_type": "logic"},
            headers=headers,
            skip_auto_create=True,
        )
        if response.status_code != 200:
            raise ResponseStatusException(
                f"Failed to delete engine {name}.", response
            )

    def suspend_engine(self, name:str):
        response = self.request(
            "suspend_engine",
            path_params={"engine_name": name, "engine_type": "logic"},
            skip_auto_create=True,
        )
        if response.status_code != 200:
            raise ResponseStatusException(
                f"Failed to suspend engine {name}.", response
            )

    def resume_engine_async(self, name:str, headers={}):
        response = self.request(
            "resume_engine",
            path_params={"engine_name": name, "engine_type": "logic"},
            headers=headers,
            skip_auto_create=True,
        )
        if response.status_code != 200:
            raise ResponseStatusException(
                f"Failed to resume engine {name}.", response
            )
        return {}

    def _poll_use_index(
        self,
        app_name: str,
        sources: Iterable[str],
        model: str,
        engine_name: str,
        engine_size: str | None = None,
        program_span_id: str | None = None,
        headers: Dict | None = None,
    ):
        """Poll use_index to prepare indices for the given sources using DirectUseIndexPoller."""
        return DirectUseIndexPoller(
            self,
            app_name=app_name,
            sources=sources,
            model=model,
            engine_name=engine_name,
            engine_size=engine_size,
            language=self.language,
            program_span_id=program_span_id,
            headers=headers,
            generation=self.generation,
        ).poll()

    def maybe_poll_use_index(
        self,
        app_name: str,
        sources: Iterable[str],
        model: str,
        engine_name: str,
        engine_size: str | None = None,
        program_span_id: str | None = None,
        headers: Dict | None = None,
    ):
        """Only call poll() if there are sources to process and cache is not valid."""
        sources_list = list(sources)
        self.database = model
        if sources_list:
            poller = DirectUseIndexPoller(
                self,
                app_name=app_name,
                sources=sources_list,
                model=model,
                engine_name=engine_name,
                engine_size=engine_size,
                language=self.language,
                program_span_id=program_span_id,
                headers=headers,
                generation=self.generation,
            )
            # If cache is valid (data freshness has not expired), skip polling
            if poller.cache.is_valid():
                cached_sources = len(poller.cache.sources)
                total_sources = len(sources_list)
                cached_timestamp = poller.cache._metadata.get("cachedIndices", {}).get(poller.cache.key, {}).get("last_use_index_update_on", "")

                message = f"Using cached data for {cached_sources}/{total_sources} data streams"
                if cached_timestamp:
                    print(f"\n{message} (cached at {cached_timestamp})\n")
                else:
                    print(f"\n{message}\n")
            else:
                return poller.poll()




import warnings

from v0.relationalai.semantics.devtools.benchmark_lqp import benchmark_lqp

__all__ = ["benchmark_lqp"]

warnings.warn(
    "relationalai.early_access.devtools.benchmark_lqp is deprecated, "
    "Please migrate to relationalai.semantics.devtools.benchmark_lqp",
    DeprecationWarning,
    stacklevel=2,
)


import warnings

from v0.relationalai.semantics.metamodel.ir import (sanitize, node_to_string, value_to_string, Task, Var, Logical,
                                                 ScalarType, Node, Capability, ListType, _global_id)

__all__ = ["sanitize", "node_to_string", "value_to_string", "Task", "Var", "Logical", "ScalarType", "Node",
           "Capability", "ListType", "_global_id"]

warnings.warn(
    "relationalai.early_access.metamodel.ir is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.ir",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations
from dataclasses import dataclass
from typing import cast, Optional, TypeVar
from typing import Tuple

from v0.relationalai.semantics.metamodel import builtins, ir, factory as f, helpers
from v0.relationalai.semantics.metamodel.compiler import Pass, group_tasks
from v0.relationalai.semantics.metamodel.util import NameCache, OrderedSet, ordered_set
from v0.relationalai.semantics.metamodel import dependency
from v0.relationalai.semantics.metamodel.typer.typer import to_type

class Flatten(Pass):
    """
    Traverses the model's root to flatten it as much as possible. The result of this pass is
    a Logical root where all nested tasks that represent a rule in Rel are extracted to the
    top level.

    - nested logical with updates becomes a top-level logical (a rule)

    From:
        Logical
            Logical
                lookup1   <- scope is spread
                Logical
                    lookup2
                    derive foo
                Logical
                    lookup3
                    derive bar
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive foo
            Logical
                lookup1
                lookup3
                derive bar

    - nested logical with aggregates becomes a top-level logical (a rule representing an aggregation)

    From:
        Logical
            Logical
                lookup1
                Logical
                    lookup2
                    aggregate1
                Logical
                    lookup3
                    aggregate2
                output
    To:
        Logical
            Logical
                lookup1
                lookup2
                aggregate1
                derive tmp1
            Logical
                lookup1
                lookup3
                aggregate2
                derive tmp2
            Logical
                lookup1
                lookup tmp1
                lookup tmp2
                output

    - a union becomes a top-level logical for each branch, writing into a temporary relation,
    and a lookup from that relation.

    From:
        Logical
            Logical
                Union
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                lookup2
                derive tmp1
            Logical
                lookup tmp1
                output

    - a match becomes a top-level logical for each branch, each writing into its own temporary
    relation and a lookup from the last relation. The top-level logical for a branch derives
    into the temporary relation negating the previous branch:

    From:
        Logical
            Logical
                Match
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                Union            <- tmp1() or (not temp1() and lookup2())
                    lookup tmp1
                    Logical
                        Not
                            lookup tmp1
                        lookup2
                        derive tmp2
            Logical
                lookup tmp2
                output
    """

    def __init__(self, use_sql: bool=False):
        super().__init__()
        self.name_cache = NameCache(start_from_one=True)
        self._use_sql = use_sql


    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # create the context
        ctx = Flatten.Context(model, options)

        # rewrite the root
        result = self.handle(model.root, ctx)

        # the new body contains the extracted top level logicals and maybe the rewritten root
        body = ctx.rewrite_ctx.top_level if result.replacement is None else ctx.rewrite_ctx.top_level + [result.replacement]

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen(),
            model.types,
            ir.Logical(model.root.engine, tuple(), tuple(body))
        )

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    class Context():
        def __init__(self, model: ir.Model, options: dict):
            self.rewrite_ctx = helpers.RewriteContext()
            self.info = dependency.analyze(model.root)
            self.options = options
            # when we extract a nested logical (due to agg, match, union, etc), we also
            # extract together its dependencies, and we leave behind a different nested
            # logical as a reference, with a lookup to the extracted relation. This dict
            # keeps a map from the reference logical id to the set of tasks that where
            # extracted for it. This is used to optimize the resulting code, since we can
            # assume that the reference logical already includes those tasks.
            self.included: dict[int, OrderedSet[ir.Task]] = dict()
            # These are extra tasks to be added to the body of a logical when it is
            # extracted as a top-level logical. For example, when handling a match, we add
            # a couple of tasks here in case the branches have updates or aggregations that
            # cause the branch to be extracted, so that they include these tasks.
            self.extra_tasks = []


    @dataclass
    class HandleResult():
        replacement: Optional[ir.Task]

    def handle(self, task: ir.Task, ctx: Context) -> Flatten.HandleResult:
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Union) and (task.hoisted or self._use_sql):
            # Only flatten Unions which hoist variables. If there are no hoisted variables,
            # then the Union acts as a filter, and it can be inefficient to flatten it.
            #
            # However, for the SQL backend, we always need to flatten Unions for correct SQL
            # generation.
            return self.handle_union(task, ctx)
        elif isinstance(task, ir.Match):
            return self.handle_match(task, ctx)
        elif isinstance(task, ir.Require):
            return self.handle_require(task, ctx)
        elif isinstance(task, ir.Not):
            return self.handle_not(task, ctx)
        else:
            return Flatten.HandleResult(task)


    def handle_logical(self, task: ir.Logical, ctx: Context):
        # keep track of what's the result of handling nested composites
        composites = group_tasks(task.body, {
            "composites": helpers.COMPOSITES
        })["composites"]
        all_composites_removed = len(composites) > 0

        # recursively handle children, collecting the replacements in the body
        body:OrderedSet[ir.Task] = ordered_set()
        for child in task.body:
            result = self.handle(child, ctx)
            if result.replacement is not None:
                ctx.info.replaced(child, result.replacement)
                extend_body(body, result.replacement)
                # nested composite was not completely removed
                if child in composites:
                    all_composites_removed = False

        # Filter out empty logicals from the body
        new_body:OrderedSet[ir.Task] = ordered_set()
        for b in body:
            if isinstance(b, ir.Logical):
                if not b.body:
                    # empty logical, skip
                    continue
            new_body.add(b)

        body = new_body

        # all children were extracted or all composites were removed without any effects
        # left and no outputs (so no way for outer dependencies), drop this logical
        if not body or (all_composites_removed and not any([isinstance(t, helpers.EFFECTS) for t in body]) and not ctx.info.task_outputs(task)):
            return Flatten.HandleResult(None)

        # now process the rewritten body
        groups = group_tasks(body.get_list(), {
            "outputs": ir.Output,
            "updates": ir.Update,
            "aggregates": ir.Aggregate,
            "ranks": ir.Rank,
        })

        # If there are outputs, flatten each into its own top-level rule, along with its
        # dependencies.
        if groups["outputs"]:
            if self._use_sql:
                ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))
                return Flatten.HandleResult(None)

            # Analyze the dependencies in the newly rewritten body
            new_logical = ir.Logical(task.engine, task.hoisted, tuple(body))
            info = dependency.analyze(new_logical)

            for output in groups["outputs"]:
                assert(isinstance(output, ir.Output))
                new_body = info.task_dependencies(output)
                new_body.update(ctx.extra_tasks)
                new_body.add(output)
                ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(new_body), task.annotations))

            return Flatten.HandleResult(None)

        # if there are updates, extract as a new top level rule
        if groups["updates"]:
            # add task dependencies to the body
            body.prefix(ctx.info.task_dependencies(task))
            # potentially add context extra tasks
            body.update(ctx.extra_tasks)
            ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))
            return Flatten.HandleResult(None)

        if groups["aggregates"]:
            if len(groups["aggregates"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            agg = cast(ir.Aggregate, groups["aggregates"].some())

            # add agg dependencies to the body
            body.prefix(ctx.info.task_dependencies(agg))

            # extract a new logical for the aggregate, exposing aggregate group-by and results
            exposed_vars = OrderedSet.from_iterable(list(agg.group) + helpers.aggregate_outputs(agg))
            name = helpers.create_task_name(self.name_cache, agg)
            connection = helpers.extract(agg, body, exposed_vars.get_list(), ctx.rewrite_ctx, name)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars.get_list())], task.hoisted)
            return Flatten.HandleResult(reference)

        if groups["ranks"]:
            if len(groups["ranks"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            rank = cast(ir.Rank, groups["ranks"].some())

            # add rank dependencies to the body
            body.prefix(ctx.info.task_dependencies(rank))
            # for rank, we sort by the args, but the result includes the keys to preserve bag semantics.
            exposed_vars_raw = list(rank.projection) + list(rank.group) + list(rank.args) +[rank.result]
            # deduplicate vars
            exposed_vars = OrderedSet.from_iterable(exposed_vars_raw)
            name = helpers.create_task_name(self.name_cache, rank)
            connection = helpers.extract(rank, body, exposed_vars.get_list(), ctx.rewrite_ctx, name)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars.get_list())], task.hoisted)
            ctx.included[reference.id] = body
            return Flatten.HandleResult(reference)

        return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body)))


    def handle_match(self, match: ir.Match, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not match.tasks:
            return Flatten.HandleResult(match)

        body = ctx.info.task_dependencies(match)
        exposed_vars = self.compute_exposed_vars(match, match.tasks, ctx)

        # negation len is the number of wildcards when negating a reference
        outputs = ctx.info.task_outputs(match)
        negation_len = len(outputs) if outputs else 0

        # collect the references to branch rules so that we union in the end
        references = []

        # a negated reference to the previous branch (e.g. "not branch1(...)")
        negated_reference = None

        # number of negated references currently in context (e.g. in the 3rd branch there
        # will be 2 negated references, "not branch1(...) and not branch2(...)")
        negated_references = 0
        for branch in match.tasks:
            # if there's a previous connection, add the negation of it to extra tasks
            if negated_reference:
                ctx.extra_tasks.append(negated_reference)
                negated_references += 1

            # the connection relation to be used for this branch
            name = helpers.create_task_name(self.name_cache, branch, "_match")
            relation = helpers.create_connection_relation(branch, exposed_vars, ctx.rewrite_ctx, name)
            branch_update = f.derive(relation, exposed_vars)
            ctx.extra_tasks.append(branch_update)

            # handle the branch
            result = self.handle(branch, ctx)

            # if there's no replacement, the branch was already extracted (e.g. by an update),
            # otherwise, we need to extract the branch as a rule writing into the connection
            if result.replacement:
                ctx.info.replaced(branch, result.replacement)

                branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
                extend_body(branch_body, result.replacement)
                # potentially add context extra tasks (negations of previous branches + derive into this branch)
                branch_body.update(ctx.extra_tasks)
                # extract the body
                ctx.rewrite_ctx.top_level.append(ir.Logical(match.engine, tuple(), tuple(branch_body)))

            # cleanup the branch update from context (but keep the negations)
            ctx.extra_tasks.pop()

            # record a reference to the branch connection and adjust the negated_reference
            reference = f.lookup(relation, exposed_vars)
            negated_reference = negate(reference, negation_len)
            references.append(reference)

        # we accumulate negations, in the end we remove them all
        for x in range(negated_references):
            ctx.extra_tasks.pop()

        # return a union of the references to the branches
        return Flatten.HandleResult(f.union(references, match.hoisted))


    def handle_union(self, union: ir.Union, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not union.tasks:
            return Flatten.HandleResult(union)

        body = ctx.info.task_dependencies(union)
        exposed_vars = self.compute_exposed_vars(union, union.tasks, ctx)

        name = helpers.create_task_name(self.name_cache, union)
        relation = helpers.create_connection_relation(union, exposed_vars, ctx.rewrite_ctx, name)
        for branch in union.tasks:

            # add an update to the context in case the branch is extracted
            branch_update = f.derive(relation, exposed_vars)
            ctx.extra_tasks.append(branch_update)

            # handle the branch
            result = self.handle(branch, ctx)

            # cleanup the branch update from context
            ctx.extra_tasks.pop()

            # if there's no replacement, the branch was already extracted (e.g. by an update),
            # otherwise, we need to extract the branch as a rule writing into the connection
            if result.replacement:
                # the branch has some replacement, we have to extract it as a top-level rule
                ctx.info.replaced(branch, result.replacement)
                branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
                extend_body(branch_body, result.replacement)
                branch_body.add(branch_update)
                # extract the body
                ctx.rewrite_ctx.top_level.append(ir.Logical(union.engine, tuple(), tuple(branch_body)))

        # return a reference to the connection relation
        reference = f.logical([f.lookup(relation, exposed_vars)], exposed_vars)
        return Flatten.HandleResult(reference)


    def compute_exposed_vars(self, task: ir.Task, branches: Tuple[ir.Task, ...], ctx: Context) -> list:
        """ Helper to compute the set of exposed vars for handle_match and handle_union. """
        exposed_vars = set_union(ctx.info.task_inputs(task), ctx.info.task_outputs(task))
        if exposed_vars:
            return exposed_vars
        # no exposed vars, so we try to use the common vars across all branches as context
        common_vars = helpers.collect_vars(branches[0])
        for branch in branches:
            branch_vars = helpers.collect_vars(branch)
            for v in common_vars:
                if v not in branch_vars:
                    common_vars.remove(v)
        # TODO - we should raise an error here if the set is empty
        return common_vars.get_list()


    def handle_require(self, req: ir.Require, ctx: Context):
        # only extract the domain if it is a somewhat complex Logical and there's more than
        # one check, otherwise insert it straight into all checks
        if builtins.discharged_annotation in req.annotations:
            # remove discharged Requires
            return Flatten.HandleResult(None)
        elif builtins.declare_constraint_annotation in req.annotations:
            # leave Requires that are declared constraints
            return Flatten.HandleResult(req)
        else:
            # generate logic for remaining requires
            domain = req.domain
            if len(req.checks) > 1 and isinstance(domain, ir.Logical) and len(domain.body) > 1:
                body = OrderedSet.from_iterable(domain.body)
                vars = helpers.hoisted_vars(domain.hoisted)
                name = helpers.create_task_name(self.name_cache, req)
                connection = helpers.extract(req, body, vars, ctx.rewrite_ctx, name)
                domain = f.logical([f.lookup(connection, vars)], vars)

            for check in req.checks:
                # only generate logic for checks that have errors and not discharged
                if check.error and builtins.discharged_annotation not in check.annotations:
                    handled_check_result = self.handle(check.check, ctx)
                    if handled_check_result.replacement:
                        body = ordered_set()
                        body.add(domain)
                        body.add(ir.Not(req.engine, handled_check_result.replacement))
                        if (isinstance(check.error, ir.Logical)):
                            body.update(check.error.body)
                        else:
                            # this is more general but may trip the current splinter
                            body.add(check.error)
                        ctx.rewrite_ctx.top_level.append(ir.Logical(req.engine, tuple(), tuple(body)))

        # currently we just drop the Require, but we should keep it here and link the
        # extracted logicals to it
        return Flatten.HandleResult(None)


    def handle_not(self, task: ir.Not, ctx: Context):
        # handle the sub-task of the Not
        result = self.handle(task.task, ctx)
        # there must be a replacement
        assert(result.replacement)

        # if the sub-task was really not replaced, just return the task, otherwise a new Not
        if result.replacement is task.task:
            return Flatten.HandleResult(task)
        else:
            return Flatten.HandleResult(ir.Not(
                task.engine,
                result.replacement,
                task.annotations
            ))

#--------------------------------------------------
# Helpers
#--------------------------------------------------

T = TypeVar('T')
def set_union(s1: Optional[OrderedSet[T]], s2: Optional[OrderedSet[T]]) -> list:
    """ Return the union of s1 | s2, accounting for when any is None. """
    if s1 and s2:
        return (s1 | s2).get_list()
    if s1:
        return s1.get_list()
    if s2:
        return s2.get_list()
    return []

def negate(lookup: ir.Lookup, values: int):
    """
    Return a negation of this reference, where the last `values` arguments are to
    be replaced by wildcards (i.e. len(reference.args) - values are keys so they need
    to be bound in the Not.)
    """
    args = []
    i = 0
    last = len(lookup.args) - values
    for arg in lookup.args:
        args.append(f.wild(to_type(arg))) if i >= last else args.append(arg)
        i += 1

    return ir.Not(lookup.engine, f.lookup(lookup.relation, args))

def merge_var_list(vars: list[ir.Var], hoisted: tuple[ir.VarOrDefault, ...]) -> list[ir.VarOrDefault]:
    """ Merge vars and hoisted, making sure that hoisted vars have precedence since they may have defaults. """
    r = []
    hoisted_vars = helpers.hoisted_vars(hoisted)
    for v in vars:
        if v not in hoisted_vars:
            r.append(v)
    r.extend(hoisted)
    return r

def extend_body(body: OrderedSet[ir.Task], extra: ir.Task):
    """ Add the extra task to the body, but if the extra is a simple logical, just
    inline its subtasks. """
    if isinstance(extra, ir.Logical):
        if extra.hoisted:
            # hoists, remove things that are already in the body to avoid duplicates
            logical_body = []
            for t in extra.body:
                if t not in body:
                    logical_body.append(t)
            if len(logical_body) == len(extra.body):
                # no duplicates
                body.add(extra)
            else:
                # some duplicate, remove them
                body.add(ir.Logical(
                    extra.engine,
                    extra.hoisted,
                    tuple(logical_body)
                ))
        else:
            # no hoists, just inline
            body.update(extra.body)
    else:
        body.add(extra)



"""The RelationalAI semantics API.

This module provides the core classes and functions for defining and querying
semantic models.
"""

from pandas import DataFrame
from ..util.error import exc
from ..util.docutils import include_in_docs
from .frontend.base import (
    Alias, CoreLibrary, Concept, Data, Distinct, Expression, Field, FieldRef, Fragment, Library,
    Literal, Match, Model, New, Not, Chain, Property, Reading, Relationship, Statement, StatementAndSchema, Table, Union, Value,
    Variable,
)
# populate the core library
from .frontend import core

# include this module in documentation
__include_in_docs__ = True

    # String, Integer, Int64, Int128, Float, Decimal, Bool,
    # Date, DateTime,
    # RawSource, Hash,
    # select, where, require, define, distinct, union, data,
    # rank, asc, desc,
    # count, sum, min, max, avg, per,
    # not_

#------------------------------------------------------
# Convenience model functions
#------------------------------------------------------

def _check_model():
    if len(Model.all_models) == 0:
        exc("Missing Model", "No Model has been defined. Please create a Model before using the semantics API.")
    elif len(Model.all_models) > 1:
        exc("Ambiguous model", "Multiple Models have been defined. Please use functions on the specific Model.")
    return Model.all_models[0]

@include_in_docs
def select(*args:StatementAndSchema) -> Fragment:
    """Return a selection fragment from the active model.

    Parameters
    ----------
    *args : StatementAndSchema
        Concepts, properties, relationships fields, expressions, or other
        statements to select.

    Returns
    -------
    Fragment
        A composable query fragment representing the selection.

    Examples
    --------
    Select a literal value and materialize the result as a pandas DataFrame:

    >>> select(1).to_df()

    Combine with filters to project properties of specific entities from the
    current model:

    >>> select(Person.name, Person.age) \
    ...     .where(Person.age > 21).to_df()  # doctest: +SKIP
    """
    model = _check_model()
    return model.select(*args)

def define(*args:Statement):
    model = _check_model()
    return model.define(*args)

def where(*args:Statement):
    model = _check_model()
    return model.where(*args)

def require(*args:Statement):
    model = _check_model()
    return model.require(*args)

def union(*args:Value):
    model = _check_model()
    return model.union(*args)

def data(data: DataFrame | list[tuple] | list[dict], columns: list[str]|None = None):
    model = _check_model()
    return model.data(data, columns)

def distinct(*args:Value):
    model = _check_model()
    return model.distinct(*args)

def not_(*args:Value):
    model = _check_model()
    return model.not_(*args)

def literal(value, type: Concept|None = None) -> Literal:
    model = _check_model()
    return Literal(value, model, type)

#------------------------------------------------------
# Convenience types
#------------------------------------------------------
Any = core.Any
AnyEntity = core.AnyEntity
TypeVar = core.TypeVar
EntityTypeVar = core.EntityTypeVar
Number = core.Number
String = core.String
Integer = core.Integer
Int = core.Int
Int64 = core.Int64
Int128 = core.Int128
Float = core.Float
Decimal = core.Decimal
Bool = core.Boolean
Boolean = Bool
Date = core.Date
DateTime = core.DateTime
# RawSource = core.RawSource
# Hash = core.Hash

#------------------------------------------------------
# Convenience aggregates
# NOTE: these should be deprecated
#------------------------------------------------------

from .std import aggregates as aggs

rank = aggs.rank
limit = aggs.limit
asc = aggs.asc
desc = aggs.desc
count = aggs.count
sum = aggs.sum
min = aggs.min
max = aggs.max
avg = aggs.avg
string_join = aggs.string_join
per = aggs.per

#------------------------------------------------------
# Exports
#------------------------------------------------------

__all__ = [
    "Alias", "CoreLibrary", "Concept", "Data", "Distinct", "Expression", "Field", "FieldRef", "Fragment",
    "Library", "Literal", "Match", "Model", "New", "Not", "Chain", "Property", "Reading", "Relationship",
    "Table", "Union", "Value", "Variable", "core",

    "select", "define", "where", "require", "union", "data", "not_",
]



from v0.relationalai.early_access.dsl import Namespace
from v0.relationalai.early_access.dsl.core.types import Type
from v0.relationalai.semantics.metamodel.util import OrderedSet
from v0.relationalai.early_access.dsl.core.relations import ExportRelation, RelationSignature


class Export:

    def __init__(self, key: list[Type], target_table: str) -> None:
        self.key = key
        self.target_table = target_table
        self.columns = OrderedSet[ExportRelation]()
        self.labels = OrderedSet[str]()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def column(self, name: str, *sig: Type, functional: bool = False):
        rel_name = f'output:col{len(self.columns):03d}'
        rel = ExportRelation(
            Namespace.top,
            rel_name,
            RelationSignature(*(self.key + list(sig)), functional=functional)
        )
        self.columns.add(rel)
        self.labels.add(name)
        return rel


from v0.relationalai.semantics.metamodel import ir as meta
from v0.relationalai.semantics.metamodel import types
from v0.relationalai.semantics.lqp import ir as lqp
from v0.relationalai.semantics.lqp.constructors import mk_type, mk_value
from typing import Union

# LQP Integer type limits
INT_MIN = -2**63 # LQP INT is a 64-bit signed integer
INT_MAX = 2**63 - 1 # LQP INT is a 64-bit signed integer
INT128_MIN = -2**127
INT128_MAX = 2**127 - 1
UINT128_MIN = 0
UINT128_MAX = 2**128 - 1

PrimitiveValue = Union[
    str, int, float, lqp.UInt128Value, lqp.Int128Value, lqp.DecimalValue, lqp.DateValue,
    lqp.DateTimeValue, lqp.BooleanValue
]

def is_numeric(type: lqp.Type) -> bool:
    # UInt128 is used for hash, RowID, and entity ids so we don't include it as numeric
    return type.type_name in {
        lqp.TypeName.INT, lqp.TypeName.INT128, lqp.TypeName.FLOAT, lqp.TypeName.DECIMAL
    }

def meta_type_to_lqp(typ: meta.Type) -> lqp.Type:
    if isinstance(typ, meta.UnionType):
        # By this point, unions can only consist of entity types. In the LQP,
        # they are undifferentiated (hashes), so they all merge.
        assert all(types.is_entity_type(t) for t in typ.types), \
            f"Union type {typ} contains non-entity types: " \
            f"{[t for t in typ.types if not types.is_entity_type(t)]}"

        return mk_type(lqp.TypeName.UINT128)
    else:
        assert isinstance(typ, meta.ScalarType)
        if isinstance(typ, meta.DecimalType):
            # deal first with decimals as they may be in the builtin set as well
            return mk_type(lqp.TypeName.DECIMAL, [mk_value(typ.precision), mk_value(typ.scale)])
        elif types.is_builtin(typ):
            if typ == types.Int64:
                return mk_type(lqp.TypeName.INT)
            elif typ == types.Int128:
                return mk_type(lqp.TypeName.INT128)
            elif typ == types.Float:
                return mk_type(lqp.TypeName.FLOAT)
            elif typ == types.String:
                return mk_type(lqp.TypeName.STRING)
            elif typ == types.Date:
                return mk_type(lqp.TypeName.DATE)
            elif typ == types.DateTime:
                return mk_type(lqp.TypeName.DATETIME)
            elif typ == types.Hash:
                return mk_type(lqp.TypeName.UINT128)
            elif typ == types.RowId:
                return mk_type(lqp.TypeName.UINT128)
            elif typ == types.UInt128:
                return mk_type(lqp.TypeName.UINT128)
            elif typ == types.Bool:
                return mk_type(lqp.TypeName.BOOLEAN)
            elif typ == types.Number:
                # All types must be specified in the LQP.
                raise Exception("Number type could not be determined.")
            elif types.is_any(typ):
                # All types must be specified in the LQP.
                raise Exception("Type could not be determined.")
            else:
                raise NotImplementedError(f"Unknown builtin type: {typ.name}")
        elif types.is_entity_type(typ):
            return mk_type(lqp.TypeName.UINT128)
        else:
            # Otherwise, the type extends some other type, we use that instead
            assert len(typ.super_types) > 0, f"Type {typ} has no super types"
            assert len(typ.super_types) == 1, f"Type {typ} has multiple super types: {typ.super_types}"
            super_type = typ.super_types[0]
            assert isinstance(super_type, meta.ScalarType), f"Super type {super_type} of {typ} is not a scalar type"
            return meta_type_to_lqp(super_type)

def lqp_type_to_sql(arg: lqp.Type) -> str:
    if arg.type_name == lqp.TypeName.INT:
        return "NUMBER(38, 0)"
    elif arg.type_name == lqp.TypeName.FLOAT:
        return "FLOAT"
    elif arg.type_name == lqp.TypeName.STRING:
        return "VARCHAR"
    elif arg.type_name == lqp.TypeName.INT128:
        return "NUMBER(38, 0)"
    elif arg.type_name == lqp.TypeName.UINT128:
        return "NUMBER(38, 0)"
    elif arg.type_name == lqp.TypeName.DATE:
        return "DATE"
    elif arg.type_name == lqp.TypeName.DATETIME:
        return "DATETIME"
    elif arg.type_name == lqp.TypeName.DECIMAL:
        assert len(arg.parameters) == 2, \
            f"DECIMAL type must have 2 parameters, got {arg.parameters}"
        precision = arg.parameters[0].value
        scale = arg.parameters[1].value
        return f"NUMBER({precision}, {scale})"
    else:
        raise NotImplementedError(f"Unknown relational value type: {arg}")



import re
import warnings

from owlready2 import owl, get_ontology
from v0.relationalai.early_access.dsl.adapters.owl.model import OWLClass, OWLObjectProperty, OWLDatatypeProperty


class OWLParser:
    _inverse_pattern = re.compile(r'Inverse\((.*?)\)')
    _disjoint_pattern = re.compile(r'Not\((.*?)\)')

    def __init__(self, owl_file_path):
        self._ontology = get_ontology(owl_file_path).load()
        self._ontology_name = re.sub(r'(?<!^)(?=[A-Z])', '_', self._ontology.name).lower()
        self._prefix = f"{self._ontology.name}."
        self._classes = self._parse_classes()
        self._parse_subclasses()
        self._object_properties = self._parse_object_properties()
        self._datatype_properties = self._parse_datatype_properties()

    def ontology(self):
        return self._ontology

    def ontology_name(self):
        return self._ontology_name

    def classes(self):
        return self._classes

    def object_properties(self):
        return self._object_properties

    def datatype_properties(self):
        return self._datatype_properties

    def _parse_classes(self):
        owl_classes = {}
        for ontology_class in self._ontology.classes():
            class_name = self._clean_name(str(ontology_class))
            owl_classes[class_name] = OWLClass(class_name)
        return owl_classes

    def _parse_subclasses(self):
        for owl_class in self._ontology.classes():
            class_name = self._clean_name(str(owl_class))
            parents = set()
            children = set()
            for sub_class in list(owl_class.subclasses()):
                sub_class_name = self._clean_name(str(sub_class))
                parents.add(class_name)
                children.add(sub_class_name)
                if len(parents) > 0:
                    if self._classes[sub_class_name].parents is not None:
                        self._classes[sub_class_name].parents.update(parents)
                    else:
                        self._classes[sub_class_name].parents = parents
            if len(children) > 0:
                self._classes[class_name].children = children
            self._parse_class_partition(owl_class)
            self._parse_disjoint_classes(owl_class)
            self._parse_mandatory_constraints(owl_class)

    def _parse_object_properties(self):
        object_properties = {}
        for object_property in self._ontology.object_properties():
            object_property_name = self._clean_name(str(object_property))
            owl_object_property = OWLObjectProperty(object_property_name, [], [])
            for domain_class in object_property.domain:
                owl_object_property.domain.append(self._clean_name(str(domain_class)))
            for range_class in object_property.range:
                owl_object_property.range.append(self._clean_name(str(range_class)))
            if object_property.inverse_property is not None:
                inverse_property_name = self._clean_name(str(object_property.inverse_property))
                owl_object_property.inverse_property = inverse_property_name
            if self._is_functional(object_property):
                owl_object_property.functional = True
            if self._is_inverse_functional(object_property):
                owl_object_property.inverse_functional = True
            object_properties[object_property_name] = owl_object_property
        return object_properties

    def _parse_datatype_properties(self):
        datatype_properties = {}
        for datatype_property in self._ontology.data_properties():
            datatype_property_name = self._clean_name(str(datatype_property))
            datatype_properties[datatype_property_name] = OWLDatatypeProperty(datatype_property_name, [], [])
            for domain_class in datatype_property.domain:
                datatype_properties[datatype_property_name].domain.append(self._clean_name(str(domain_class)))
            for range_iri in datatype_property.range:
                datatype_properties[datatype_property_name].range.append(range_iri)
            if self._is_functional(datatype_property):
                datatype_properties[datatype_property_name].functional = True
        return datatype_properties

    def _parse_mandatory_constraints(self, owl_class):
        class_name = self._clean_name(str(owl_class))
        mandatory_properties = set()
        inverse_mandatory_properties = set()
        for isa in owl_class.is_a:
            if "some" in str(isa):
                match = self._inverse_pattern.search(str(isa.property))
                if match:
                    result = match.group(1)
                    inverse_mandatory_properties.add(self._clean_name(result))
                else:
                    mandatory_properties.add(self._clean_name(str(isa.property)))
        if len(mandatory_properties) > 0:
            self._classes[class_name].mandatory_properties = mandatory_properties
        if len(inverse_mandatory_properties) > 0:
            self._classes[class_name].inverse_mandatory_properties = inverse_mandatory_properties

    def _parse_class_partition(self, owl_class):
        owl_class_name = self._clean_name(str(owl_class))
        partition = set()
        for equivalent_class in owl_class.equivalent_to:
            if "&" in str(equivalent_class):
                warnings.warn(f"Unable to parse complex expression: {owl_class_name} equivalent to {owl_class.equivalent_to}. Skipping this axiom")
            elif "|" in str(equivalent_class):
                for partition_class in equivalent_class.Classes:
                    partition.add(self._clean_name(str(partition_class)))
        if len(partition) > 0:
            self._classes[owl_class_name].partition = partition

    def _parse_disjoint_classes(self, owl_class):
        owl_class_name = self._clean_name(str(owl_class))
        disjoints = set()
        for isa in owl_class.is_a:
            match = self._disjoint_pattern.search(str(isa))
            if match:
                disjoints.add(self._clean_name(match.group(1)))
        if len(disjoints) > 0:
            self._classes[owl_class_name].disjoint = disjoints

    def _clean_name(self, name):
        return name.removeprefix(self._prefix)

    @staticmethod
    def _is_functional(object_property):
        return owl.FunctionalProperty in object_property.is_a

    @staticmethod
    def _is_inverse_functional(object_property):
        return owl.InverseFunctionalProperty in object_property.is_a

    def is_partitioned_into_disjoint_subclasses(self, owl_class: OWLClass):
        if owl_class.partition is not None:
            for sub_class in owl_class.partition:
                disjoint = self._classes[sub_class].disjoint
                if disjoint is not None and owl_class.partition.difference(disjoint) != {sub_class}:
                    return False
            return True
        return False

    @staticmethod
    def is_partitioned_into_subclasses(owl_class: OWLClass):
        return owl_class.partition is not None

    @staticmethod
    def is_root_class(owl_class: OWLClass):
        return owl_class.parents is None



from v0.relationalai.errors import RAIWarning, RAIException

# -----------------------------------------------------------------------------------------
# All User-facing diagnostic messages: Warnings and Errors
# -----------------------------------------------------------------------------------------

class PathfinderParameterWarning(RAIWarning):
    def __init__(self, message: str):
        super().__init__(message)

class PathfinderSizeWarning(RAIWarning):
    def __init__(self, pattern, prune_states_only: bool):
        from v0.relationalai.experimental.pathfinder.rpq import RPQ
        assert isinstance(pattern, RPQ)

        message = f"WARNING: RPQ pattern is prohibitively large {len(pattern)} >= 50"
        if prune_states_only:
            message += "\nSkipping transitions pruning"
            message += "\nUse 'force_transition_pruning=True' to force transition pruning"
        super().__init__(message)

class PathfinderNonDeterminismWarning(RAIWarning):
    def __init__(self, pattern, a, prune_states_only: bool):
        from v0.relationalai.experimental.pathfinder.rpq import RPQ
        from v0.relationalai.experimental.pathfinder.automaton import Automaton
        assert isinstance(pattern, RPQ)
        assert isinstance(a, Automaton)

        message = "WARNING: RPQ pattern is non-deterministic\n"
        if prune_states_only:
            message += "Using 'force_transition_pruning=True' may solve this issue\n"
        message += str(pattern)
        message += "\n"
        for reason in a.determinism_report():
            message += " *" + reason + '\n'
        message += "Pathfinder may return duplicate results"
        super().__init__(message)

class PathfinderUngroundedPatternError(RAIException):
    def __init__(self, pattern):
        from v0.relationalai.experimental.pathfinder.rpq import RPQ
        assert isinstance(pattern, RPQ)

        message = "ERROR: Input RPQ pattern is ungrounded\n"
        message += "Pattern potentially admits an infinite number of single-node paths\n"
        message += str(pattern)
        super().__init__(message)

class PathfinderVoidPatternError(RAIException):
    def __init__(self, pattern):
        from v0.relationalai.experimental.pathfinder.rpq import RPQ
        assert isinstance(pattern, RPQ)

        message = "ERROR: Input RPQ pattern is void (admits no paths)\n"
        message += str(pattern)
        super().__init__(message)



import warnings

from v0.relationalai.semantics.devtools.extract_lqp import extract_rai_calls

__all__ = ["extract_rai_calls"]

warnings.warn(
    "relationalai.early_access.devtools.extract_lqp is deprecated. "
    "Please migrate to relationalai.semantics.devtools.extract_lqp",
    DeprecationWarning,
    stacklevel=2,
)


import v0.relationalai as rai
from concurrent.futures import ThreadPoolExecutor, TimeoutError, as_completed
import re
import logscope
import argparse

from datetime import datetime, timedelta
from relationalai_test_util.fixtures import make_config

log = logscope.logger()

def contains_hash_like_substring(s):
    """
    Checks whether the input string contains a hash-like substring.
    A hash-like substring is defined as a string of 8 or more hexadecimal characters.

    Args:
        s (str): The input string to check.

    Returns:
        bool: True if a hash-like substring is found, False otherwise.
    """
    hex_pattern = re.compile(r'[a-fA-F0-9]{8,}')
    return bool(hex_pattern.search(s))

def should_delete(m, users, days):
    cutoff = (datetime.now() + timedelta(days=-days)).strftime('%Y-%m-%dT%H:%M')

    # Always preserve pyrel_root_db
    if m['name'] == 'pyrel_root_db':
        return False

    return (
        m['created_by'] in users
        and contains_hash_like_substring(m['name'])
        and str(m['created_on']) < cutoff
    )

def delete_model(provider, model_name):
    try:
        provider.delete_model(model_name)
        log(model_name, "deleted")
        return True, model_name
    except Exception as e:
        return False, f"Error for model {model_name}: {e}"

def delete_more_concurrently(users, days, max_workers=10, timeout=15):
    config = make_config()
    # Disable graph index since we are directly collecting and deleting rai databases for passed users
    config.set("use_graph_index", False)
    provider = rai.Provider(config=config)
    models = [m for m in provider.list_models() if should_delete(m, users, days)]

    log(f"Deleting {len(models)} models...")

    deleted_count = 0

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = {
            executor.submit(delete_model, provider, m['name']): m['name']
            for m in models
        }

        for future in as_completed(futures):
            model_name = futures[future]
            try:
                success, result = future.result(timeout=timeout)
                if success:
                    deleted_count += 1
                else:
                    log(result)
            except TimeoutError:
                log(f"Timeout reached for model: {model_name}. Moving on.")
            except Exception as e:
                log(f"Unexpected error for model {model_name}: {e}")

    return deleted_count

def main():
    parser = argparse.ArgumentParser(description='Clean up databases')
    parser.add_argument('--users', nargs='+', required=True,
                        help='List of users whose databases to clean up')
    parser.add_argument('--days', type=int, required=True,
                        help='Delete databases older than this many days')

    args = parser.parse_args()

    log(f"Starting cleanup for users: {args.users}")
    log(f"Deleting databases older than {args.days} days")

    while d := delete_more_concurrently(args.users, args.days):
        log(f"Deleted {d} more models...")

if __name__ == '__main__':
    main()
